import base64
from threading import Thread

from autobahn.twisted.websocket import WebSocketServerProtocol, \
    WebSocketServerFactory
from jarbas_hive_mind.database.client import ClientDatabase
from jarbas_hive_mind.utils.log import LOG
from jarbas_hive_mind.utils.messagebus.message import Message
from jarbas_hive_mind.utils.messagebus.ws import WebsocketClient

author = "jarbasAI"

NAME = "JarbasMindv0.1"

users = ClientDatabase()


# protocol
class JarbasMindProtocol(WebSocketServerProtocol):
    def onConnect(self, request):

        LOG.info("Client connecting: {0}".format(request.peer))
        # validate user
        userpass_encoded = bytes(request.headers.get("authorization"),
                                 encoding="utf-8")[2:-1]
        userpass_decoded = base64.b64decode(userpass_encoded).decode("utf-8")
        name, key = userpass_decoded.split(":")
        ip = request.peer.split(":")[1]
        context = {"source": self.peer}
        self.platform = request.headers.get("platform", "unknown")

        try:
            user = users.get_client_by_api_key(key)
        except:
            LOG.info("Client provided an invalid api key")
            self.factory.mycroft_send("hive.client.connection.error",
                                      {"error": "invalid api key",
                                       "ip": ip,
                                       "api_key": key,
                                       "platform": self.platform},
                                      context)
            raise ValueError("Invalid API key")
        # send message to internal mycroft bus
        data = {"ip": ip, "headers": request.headers}
        self.blacklist = users.get_blacklist_by_api_key(key)
        self.factory.mycroft_send("hive.client.connect", data, context)
        # return a pair with WS protocol spoken (or None for any) and
        # custom headers to send in initial WS opening handshake HTTP response
        headers = {"server": NAME}
        return (None, headers)

    def onOpen(self):
        """
       Connection from client is opened. Fires after opening
       websockets handshake has been completed and we can send
       and receive messages.

       Register client in factory, so that it is able to track it.
       """
        self.factory.register_client(self, self.platform)
        LOG.info("WebSocket connection open.")

    def onMessage(self, payload, isBinary):
        if isBinary:
            LOG.info(
                "Binary message received: {0} bytes".format(len(payload)))
        else:
            LOG.info(
                "Text message received: {0}".format(payload.decode('utf8')))

        self.factory.process_message(self, payload, isBinary)

    def onClose(self, wasClean, code, reason):
        self.factory.unregister_client(self, reason=u"connection closed")
        LOG.info("WebSocket connection closed: {0}".format(reason))
        ip = self.peer.split(":")[1]
        data = {"ip": ip, "code": code, "reason": "connection closed",
                "wasClean": wasClean}
        context = {"source": self.peer}
        self.factory.mycroft_send("hive.client.disconnect", data, context)

    def connectionLost(self, reason):
        """
       Client lost connection, either disconnected or some error.
       Remove client from list of tracked connections.
       """
        self.factory.unregister_client(self, reason=u"connection lost")
        LOG.info("WebSocket connection lost: {0}".format(reason))
        ip = self.peer.split(":")[1]
        data = {"ip": ip, "reason": "connection lost"}
        context = {"source": self.peer}
        self.factory.mycroft_send("hive.client.disconnect", data, context)


# server internals
class JarbasMind(WebSocketServerFactory):
    def __init__(self, bus=None, *args, **kwargs):
        super(JarbasMind, self).__init__(*args, **kwargs)
        # list of clients
        self.clients = {}
        # ip block policy
        self.ip_list = []
        self.blacklist = True  # if False, ip_list is a whitelist
        # mycroft_ws
        self.bus = bus
        self.bus_daemon = None
        self.create_mycroft_connection()

    def mycroft_send(self, type, data=None, context=None):
        data = data or {}
        context = context or {}
        if "client_name" not in context:
            context["client_name"] = NAME
        self.bus.emit(Message(type, data, context))

    def connect_to_mycroft(self):
        self.bus.run_forever()

    def create_mycroft_connection(self):
        # connect to mycroft internal websocket
        self.bus = self.bus or WebsocketClient()
        self.register_mycroft_messages()
        self.bus_daemon = Thread(target=self.connect_to_mycroft)
        self.bus_daemon.setDaemon(True)
        self.bus_daemon.start()

    def register_mycroft_messages(self):
        # HACK, TODO find why failing
        # self.bus.on('message', self.handle_message)

        def wrapper(cl, message):
            message = Message.deserialize(message)
            self.handle_message(message)

        self.bus.on("message", self.handle_message)
        # self.bus.client.on_message = wrapper

        self.bus.on('hive.client.broadcast', self.handle_broadcast)
        self.bus.on('hive.client.send', self.handle_send)

    # websocket handlers
    def register_client(self, client, platform=None):
        """
       Add client to list of managed connections.
       """
        platform = platform or "unknown"
        LOG.info("registering client: " + str(client.peer))
        t, ip, sock = client.peer.split(":")
        # see if ip address is blacklisted
        if ip in self.ip_list and self.blacklist:
            LOG.warning("Blacklisted ip tried to connect: " + ip)
            self.unregister_client(client, reason=u"Blacklisted ip")
            return
        # see if ip address is whitelisted
        elif ip not in self.ip_list and not self.blacklist:
            LOG.warning("Unknown ip tried to connect: " + ip)
            #  if not whitelisted kick
            self.unregister_client(client, reason=u"Unknown ip")
            return
        self.clients[client.peer] = {"object": client,
                                     "status": "connected",
                                     "platform": platform}

    def unregister_client(self, client, code=3078,
                          reason=u"unregister client request"):
        """
       Remove client from list of managed connections.
       """
        LOG.info("deregistering client: " + str(client.peer))
        if client.peer in self.clients.keys():
            client_data = self.clients[client.peer] or {}
            j, ip, sock_num = client.peer.split(":")
            context = {"user": client_data.get("names", ["unknown_user"])[0],
                       "source": client.peer}
            self.bus.emit(
                Message("hive.client.disconnect",
                        {"reason": reason, "ip": ip, "sock": sock_num},
                        context))
            client.sendClose(code, reason)
            self.clients.pop(client.peer)

    def process_message(self, client, payload, isBinary):
        """
       Process message from client, decide what to do internally here
       """
        LOG.info("processing message from client: " + str(client.peer))
        client_data = self.clients[client.peer]
        client_protocol, ip, sock_num = client.peer.split(":")

        if isBinary:
            # TODO receive files
            pass
        else:
            # add context for this message
            payload = payload.decode("utf-8")
            message = Message.deserialize(payload)
            message.context["source"] = client.peer
            message.context["destination"] = "skills"
            if "platform" not in message.context:
                message.context["platform"] = client_data.get("platform",
                                                              "unknown")

            # messages/skills/intents per user
            if message.type in client.blacklist.get("messages", []):
                LOG.warning(client.peer + " sent a blacklisted message " \
                                          "type: " + message.type)
                return
            # TODO check intent / skill that will trigger

            # send client message to internal mycroft bus
            self.mycroft_send(message.type, message.data, message.context)

    # mycroft handlers
    def handle_send(self, message):
        # send message to client
        msg = message.data.get("payload")
        is_file = message.data.get("isBinary")
        peer = message.data.get("peer")
        if is_file:
            # TODO send file
            pass
        elif peer in self.clients:
            # send message to client
            client = self.clients[peer]
            payload = Message.serialize(msg)
            client.sendMessage(payload, False)
        else:
            LOG.error("That client is not connected")
            self.mycroft_send("hive.client.send.error",
                              {"error": "That client is not connected",
                               "peer": peer}, message.context)

    def handle_broadcast(self, message):
        # send message to all clients
        msg = message.data.get("payload")
        is_file = message.data.get("isBinary")
        if is_file:
            # TODO send file
            pass
        else:
            # send message to all clients
            server_msg = Message.serialize(msg)
            self.broadcast(server_msg)

    def handle_message(self, message=None):
        # forward internal messages to clients if they are the target
        message = Message.deserialize(message)
        if message.type == "complete_intent_failure":
            message.type = "hive.complete_intent_failure"
        message.context = message.context or {}
        peer = message.context.get("destination")
        if peer and peer in self.clients:
            client_data = self.clients[peer] or {}
            client = client_data.get("object")
            message = message.serialize()
            client.sendMessage(bytes(message, encoding="utf-8"),
                               False)

    def shutdown(self):
        self.bus.remove('message', self.handle_message)
        self.bus.remove('hive.client.broadcast', self.handle_broadcast)
        self.bus.remove('hive.client.send', self.handle_send)



#!/usr/bin/env python
import json
import logging
import os
from subprocess import check_output
from threading import Thread

import tornado.httpserver
import tornado.options
import tornado.web
import tornado.websocket
from autobahn.twisted.websocket import WebSocketClientFactory, \
    WebSocketClientProtocol
from twisted.internet.protocol import ReconnectingClientFactory

platform = "JarbasWebChatTerminalv0.1"
logger = logging.getLogger(platform)
logger.setLevel("INFO")

ip = check_output([b'hostname', b'--all-ip-addresses']) \
    .decode("utf-8").split()[-1]
port = 8286
clients = {}
lang = "en-us"


class MainHandler(tornado.web.RequestHandler):
    def get(self):
        self.render('index.html', ip=ip, port=port)


class StaticFileHandler(tornado.web.RequestHandler):
    def get(self):
        self.render('js/app.js')


class JarbasWebChatTerminalProtocol(WebSocketClientProtocol):
    chat = None
    config = {"port": port,
              "ssl": False}

    def onConnect(self, response):
        logger.info("[INFO] Server connected: {0}".format(response.peer))

    def onOpen(self):
        logger.info("[INFO] WebSocket connection open. ")
        self.factory.client = self
        self.factory.status = "connected"
        self.webchat = Thread(target=self.serve_chat)
        self.webchat.setDaemon(True)
        self.webchat.start()

    def serve_chat(self):
        WebSocketHandler.server = self
        port = self.config.get("port", 8286)
        cert = self.config.get("cert_file", "")
        key = self.config.get("key_file", "")

        tornado.options.parse_command_line()

        routes = [
            tornado.web.url(r"/", MainHandler, name="main"),
            tornado.web.url(r"/static/(.*)", tornado.web.StaticFileHandler,
                            {'path': './'}),
            tornado.web.url(r"/ws", WebSocketHandler)
        ]

        settings = {
            "debug": True,
            "template_path": os.path.join(os.path.dirname(__file__),
                                          "templates"),
            "static_path": os.path.join(os.path.dirname(__file__), "static"),
        }

        application = tornado.web.Application(routes, **settings)
        if self.config.get("ssl", False):
            httpServer = tornado.httpserver.HTTPServer(application,
                                                       ssl_options={
                                                           "certfile": cert,
                                                           "keyfile": key
                                                       })
            url = "https://" + str(ip) + ":" + str(port)
        else:
            httpServer = tornado.httpserver.HTTPServer(application)
            url = "http://" + str(ip) + ":" + str(port)

        tornado.options.parse_command_line()

        logger.info("[INFO] Access from web browser " + url)

        httpServer.listen(port)
        tornado.ioloop.IOLoop.instance().start()

    def onMessage(self, payload, isBinary):
        if not isBinary:
            payload = payload.decode("utf-8")
            msg = json.loads(payload)
            if msg.get("type", "") == "speak":
                utterance = msg["data"]["utterance"]
                logger.info("[OUTPUT] " + utterance)
                if self.chat is not None:
                    self.chat.write_message(utterance)
            elif msg.get("type", "") == "hive.complete_intent_failure":
                logger.error("[ERROR] complete intent failure")
        else:
            pass

    def onClose(self, wasClean, code, reason):
        logger.info("[INFO] WebSocket connection closed: {0}".format(reason))
        if "Internalservererror:InvalidAPIkey" in reason:
            logger.error("[ERROR] invalid user:key provided")
            # TODO show something in web ui?
            raise ConnectionAbortedError("invalid user:key provided")


class WebSocketHandler(tornado.websocket.WebSocketHandler):
    peer = "unknown"
    server = None

    def open(self):
        logger.info('[INFO] Client IP: ' + self.request.remote_ip)
        self.peer = self.request.remote_ip
        clients[self.peer] = self
        self.write_message("Welcome to Jarbas Web Client")
        self.server.chat = self

    def on_message(self, message):
        utterance = message.strip()

        if utterance:
            logger.info("[INFO] Utterance : " + utterance)
            data = {"utterances": [utterance], "lang": lang}
            context = {"source": self.peer,
                       "destination": "hive_mind",
                       "platform": platform}
            msg = {"data": data,
                   "type": "recognizer_loop:utterance",
                   "context": context}
            msg = json.dumps(msg)
            msg = bytes(msg, encoding="utf-8")
            self.server.sendMessage(msg, False)

    def on_close(self):
        global clients
        try:
            clients.pop(self.peer)
        except:
            pass


class JarbasWebChatTerminal(WebSocketClientFactory,
                            ReconnectingClientFactory):
    protocol = JarbasWebChatTerminalProtocol

    def __init__(self, *args, **kwargs):
        super(JarbasWebChatTerminal, self).__init__(*args, **kwargs)
        self.status = "disconnected"
        self.client = None

    # websocket handlers
    def clientConnectionFailed(self, connector, reason):
        logger.info("[INFO] Client connection failed: " + str(
            reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)

    def clientConnectionLost(self, connector, reason):
        logger.info("[INFO] Client connection lost: " + str(
            reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)





import base64
from twisted.internet import reactor, ssl
from jarbas_hive_mind.drones.simple import JarbasDrone, JarbasDroneProtocol, platform


def connect_to_hivemind(host="127.0.0.1", port=5678, name="Jarbas Drone",
                        key="drone_key", useragent=platform, bus=None):
    authorization = bytes(name + ":" + key, encoding="utf-8")
    usernamePasswordDecoded = authorization
    key = base64.b64encode(usernamePasswordDecoded)
    headers = {'authorization': key}
    address = u"wss://" + host + u":" + str(port)
    factory = JarbasDrone(bus=bus, headers=headers,
                          useragent=useragent)
    factory.protocol = JarbasDroneProtocol
    contextFactory = ssl.ClientContextFactory()
    reactor.connectSSL(host, port, factory, contextFactory)
    reactor.run()


if __name__ == '__main__':
    # TODO arg parse
    connect_to_hivemind()



import json
import logging
import socket
import sys
from threading import Thread

from autobahn.twisted.websocket import WebSocketClientFactory, \
    WebSocketClientProtocol
from twisted.internet.protocol import ReconnectingClientFactory

platform = "JarbasTwitchBridgev0.1"
logger = logging.getLogger(platform)
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel("INFO")


class JarbasTwitchBridgeProtocol(WebSocketClientProtocol):
    # Set all the variables necessary to connect to Twitch IRC
    HOST = "irc.twitch.tv"
    NICK = "jarbas_bot"
    PORT = 6667
    PASS = "oauth:GET_YOURS"
    CHANNELNAME = "jarbasai"

    def onConnect(self, response):
        logger.info("[INFO] " + "Server connected: {0}".format(response.peer))
        self.factory.client = self
        self.factory.status = "connected"

    def onOpen(self):
        logger.info("[INFO] " + "WebSocket connection open. ")
        self.NICK = self.factory.username
        self.PASS = self.factory.oauth
        self.connect_to_twitch()
        self.input_loop = Thread(target=self.twitch_loop)
        self.input_loop.setDaemon(True)
        self.input_loop.start()

    def connect_to_twitch(self):
        # Connecting to Twitch IRC by passing credentials and joining a certain channel
        self.s = socket.socket()
        self.s.connect((self.HOST, self.PORT))
        self.s.send(b"PASS " + bytes(self.PASS, encoding="utf-8") + b"\r\n")
        self.s.send(b"NICK " + bytes(self.NICK, encoding="utf-8") + b"\r\n")
        self.s.send(
            b"JOIN #" + bytes(self.CHANNELNAME, encoding="utf-8") + b" \r\n")

    # Method for sending a message
    def twitch_send(self, message):
        msg = bytes("PRIVMSG #" + self.CHANNELNAME + " :" + message +
                    "\r\n", encoding="utf-8")
        self.s.send(msg)

    def onMessage(self, payload, isBinary):
        if not isBinary:
            payload = payload.decode("utf-8")
            msg = json.loads(payload)
            if msg.get("type", "") == "speak":
                utterance = msg["data"]["utterance"]
                user = msg.get("context", {}).get("user")
                logger.info("[OUTPUT] " + utterance)
                if user:
                    utterance += " " + user
                self.twitch_send(utterance)
            if msg.get("type", "") == "server.complete_intent_failure":
                utterance = "does not compute"
                logger.error("Output: complete_intent_failure")
                logger.info("[OUTPUT] " + utterance)
                self.twitch_send(utterance)
        else:
            pass

    def onClose(self, wasClean, code, reason):
        logger.info("[INFO] " + "WebSocket connection closed: {0}".format(
            reason))
        self.factory.client = None
        self.factory.status = "disconnected"
        if "Internalservererror:InvalidAPIkey" in reason:
            logger.error("[ERROR] invalid user:key provided")
            raise ConnectionAbortedError("invalid user:key provided")

    def twitch_loop(self):

        readbuffer = ""
        MODT = False

        while True:
            readbuffer = readbuffer + self.s.recv(1024).decode("utf-8")
            temp = readbuffer.split("\n")
            readbuffer = temp.pop()

            for line in temp:
                # Splits the given string so we can work with it better
                parts = line.split(":")
                # Checks whether the message is PING because its a method of Twitch to check if you're afk
                if (parts[0].strip() == "PING"):
                    msg = bytes("PONG %s\r\n" % parts[1], encoding="utf-8")
                    self.s.send(msg)
                else:
                    # Only works after twitch is done announcing stuff (MODT = Message of the day)
                    if not MODT:
                        for l in parts:
                            if "End of /NAMES list" in l:
                                MODT = True
                        continue
                    if "QUIT" not in parts[1] and "JOIN" not in parts[
                        1] and "PART" not in parts[1]:
                        try:
                            # Sets the message variable to the actual message sent
                            message = parts[2][:len(parts[2]) - 1]
                        except:
                            continue
                        # Sets the username variable to the actual username
                        usernamesplit = parts[1].split("!")
                        username = usernamesplit[0]
                        if "@" + self.NICK.lower() in message.lower():
                            message = message.replace("@" + self.NICK, "")
                            self.handle_twitch_mention(message, username)
                        else:
                            self.handle_twitch_message(message, username)

    def handle_twitch_message(self, message, sender):
        logger.info("[INFO] " + sender + " said: " + message)

    def handle_twitch_mention(self, message, sender):
        logger.info("[INFO] " + sender + " said to you: " + message)
        msg = {"data": {"utterances": [message],
                        "lang": "en-us"},
               "type": "recognizer_loop:utterance",
               "context": {"source": "twitch",
                           "destination": "hive_mind",
                           "platform": platform,
                           "user": sender}}
        msg = json.dumps(msg)
        msg = bytes(msg, encoding="utf-8")
        self.sendMessage(msg, False)


class JarbasTwitchBridge(WebSocketClientFactory,
                         ReconnectingClientFactory):
    protocol = JarbasTwitchBridgeProtocol

    def __init__(self, oauth, channel, username="Jarbas_BOT", *args, **kwargs):
        super(JarbasTwitchBridge, self).__init__(*args, **kwargs)
        self.status = "disconnected"
        self.client = None
        self.oauth = oauth
        self.channel = channel
        self.username = username

    # websocket handlers
    def clientConnectionFailed(self, connector, reason):
        logger.info(
            "Client connection failed: " + str(reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)

    def clientConnectionLost(self, connector, reason):
        logger.info(
            "Client connection lost: " + str(reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)




from jarbas_hive_mind.terminals.basic_web import RemiTerminal, start


def connect_to_hivemind_server(host="127.0.0.1", port=5678,
                               name="Jarbas Remi Terminal",
                               key="remi_key", remi_host='127.0.0.1',
                               remi_port=8171):
    RemiTerminal.host = host
    RemiTerminal.port = port
    RemiTerminal.name = name
    RemiTerminal.key = key
    start(RemiTerminal, address=remi_host, port=remi_port,
          multiple_instance=True,
          enable_file_cache=True, update_interval=0.1, start_browser=False)


def connect_to_hivemind_standalone(host="127.0.0.1", port=5678,
                                   name="Jarbas Remi Terminal",
                                   key="remi_key"):
    RemiTerminal.host = host
    RemiTerminal.port = port
    RemiTerminal.name = name
    RemiTerminal.key = key
    start(RemiTerminal, standalone=True)


if __name__ == "__main__":
    # TODO argparse
    connect_to_hivemind_server()


from sqlalchemy.ext.declarative import declarative_base

__author__ = "JarbasAI"


Base = declarative_base()


def model_to_dict(obj):
    serialized_data = {c.key: getattr(obj, c.key) for c in obj.__table__.columns}
    return serialized_data


def props(cls):
    return [i for i in cls.__dict__.keys() if i[:1] != '_']




import json
import logging
import sys
from threading import Thread
from time import sleep

from autobahn.twisted.websocket import WebSocketClientFactory, \
    WebSocketClientProtocol
from twisted.internet.protocol import ReconnectingClientFactory

platform = "JarbasHackChatBridgev0.2"
logger = logging.getLogger(platform)
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel("INFO")


class HackChat:
    """A library to connect to https://hack.chat.
    <on_message> is <list> of callback functions to receive data from
    https://hack.chat. Add your callback functions to this attribute.
    e.g., on_message += [my_callback]
    The callback function should have 3 parameters, the first for the
    <HackChat> object, the second for the message someone sent and the
    third for the nickname of the sender of the message.

    https://github.com/gkbrk/hackchat
    """

    def __init__(self, nick, channel="programming", debug=True):
        """Connects to a channel on https://hack.chat.
        Keyword arguments:
        nick -- <str>; the nickname to use upon joining the channel
        channel -- <str>; the channel to connect to on https://hack.chat
        """

        import websocket
        self.debug = debug
        self.nick = nick
        self.channel = channel
        self.online_users = []
        self.on_message = []
        self.on_join = []
        self.on_open = []
        self.on_leave = []
        self.ws = websocket.create_connection("wss://hack.chat/chat-ws")
        self._send_packet({"cmd": "join", "channel": channel, "nick": nick})
        Thread(target=self._ping_thread).start()

    def send_message(self, msg):
        """Sends a message on the channel."""
        self._send_packet({"cmd": "chat", "text": msg})

    def _send_packet(self, packet):
        """Sends <packet> (<dict>) to https://hack.chat."""
        encoded = json.dumps(packet)
        self.ws.send(encoded)

    def run(self):
        """Sends data to the callback functions."""
        while True:
            result = json.loads(self.ws.recv())
            if self.debug:
                print(result)

            if result["cmd"] == "chat" and not result["nick"] == self.nick:
                for handler in list(self.on_message):
                    handler(self, result["text"], result["nick"])
            elif result["cmd"] == "onlineAdd":
                self.online_users.append(result["nick"])
                for handler in list(self.on_join):
                    handler(self, result["nick"])
            elif result["cmd"] == "onlineRemove":
                self.online_users.remove(result["nick"])
                for handler in list(self.on_leave):
                    handler(self, result["nick"])
            elif result["cmd"] == "onlineSet":
                for nick in result["nicks"]:
                    self.online_users.append(nick)
                for handler in list(self.on_open):
                    handler(self, result["nicks"])

    def _ping_thread(self):
        """Retains the websocket connection."""
        while self.ws.connected:
            self._send_packet({"cmd": "ping"})
            sleep(60)


class JarbasHackChatBridgeProtocol(WebSocketClientProtocol):
    hackchat = None
    chat_thread = None
    username = "Jarbas_BOT"
    hack_chat_channel = "JarbasAI"

    @property
    def online_users(self):
        try:
            return self.hackchat.online_users
        except:
            return []

    def start_hack_chat(self):
        self.hackchat = HackChat(self.username, self.hack_chat_channel)
        self.hackchat.on_message += [self.on_hack_message]
        self.hackchat.on_join += [self.on_hack_join]
        self.hackchat.on_open += [self.on_hack_open]
        self.hackchat.on_leave += [self.on_hack_leave]
        self.hackchat.run()

    def on_hack_open(self, connector, users):
        if len(users) == 1:
            self.hackchat.send_message("This channel belongs to me")
        else:
            self.hackchat.send_message("I see {} online users"
                                       .format(len(users) - 1))

    def on_hack_join(self, connector, user):
        self.hackchat.send_message("Hello @{}".format(user))

    def on_hack_leave(self, connector, user):
        self.hackchat.send_message("@{} vanished from cyberspace".format(user))

    def on_hack_message(self, connector, message, user):
        utterance = message.lower().strip()
        if "@" + self.username.lower() in utterance:
            utterance = utterance.replace("@" + self.username.lower(), "")
            msg = {"data": {"utterances": [utterance], "lang": "en-us"},
                   "type": "recognizer_loop:utterance",
                   "context": {
                       "source": self.peer,
                       "destination": "hive_mind",
                       "platform": platform,
                       "hack_chat_nick": user,
                       "user": user}}
            msg = json.dumps(msg)
            msg = bytes(msg, encoding="utf-8")
            self.sendMessage(msg, False)

    def onConnect(self, response):
        logger.info("[INFO] " + "Server connected: {0}".format(response.peer))
        self.factory.client = self
        self.factory.status = "connected"
        self.username = self.factory.username
        self.hack_chat_channel = self.factory.channel
        logger.info("[INFO] " + "Channel: {0}".format(self.hack_chat_channel))
        logger.info("[INFO] " + "Username: {0}".format(self.username))

    def onOpen(self):
        logger.info("[INFO] " + "WebSocket connection open. ")
        self.chat_thread = Thread(target=self.start_hack_chat)
        self.chat_thread.setDaemon(True)
        self.chat_thread.start()

    def onMessage(self, payload, isBinary):
        if not isBinary:
            payload = payload.decode("utf-8")
            msg = json.loads(payload)
            user = msg.get("context", {}).get("hack_chat_nick", "")
            if user not in self.online_users:
                logger.error("[ERROR] " + "invalid hack chat user: " + user)
                return
            utterance = ""
            if msg.get("type", "") == "speak":
                utterance = msg["data"]["utterance"]
            elif msg.get("type", "") == "server.complete_intent_failure":
                utterance = "i can't answer that yet"

            if utterance:
                utterance = "@{} , ".format(user) + utterance
                self.hackchat.send_message(utterance)
        else:
            pass

    def onClose(self, wasClean, code, reason):
        logger.info(
            "[INFO] " + "WebSocket connection closed: {0}".format(reason))
        if self.chat_thread:
            self.chat_thread.join(0)
        self.hackchat = None
        self.factory.client = None
        self.factory.status = "disconnected"
        if "Internalservererror:InvalidAPIkey" in reason:
            logger.error("[ERROR] invalid user:key provided")
            raise ConnectionAbortedError("invalid user:key provided")


class JarbasHackChatBridge(WebSocketClientFactory,
                           ReconnectingClientFactory):
    protocol = JarbasHackChatBridgeProtocol

    def __init__(self, username, channel, *args, **kwargs):
        super(JarbasHackChatBridge, self).__init__(*args, **kwargs)
        self.status = "disconnected"
        self.client = None
        self.username = username
        self.channel = channel

    # websocket handlers
    def clientConnectionFailed(self, connector, reason):
        logger.info("[INFO] " + "Client connection failed: " + str(
            reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)

    def clientConnectionLost(self, connector, reason):
        logger.info("[INFO] " + "Client connection lost: " + str(
            reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)





# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import re

from lingua_franca.parse import normalize


class Message:
    """Holds and manipulates data sent over the websocket

        Message objects will be used to send information back and forth
        between processes of Mycroft.

    Attributes:
        type (str): type of data sent within the message.
        data (dict): data sent within the message
        context: info about the message not part of data such as source,
            destination or domain.
    """

    def __init__(self, type, data=None, context=None):
        """Used to construct a message object

        Message objects will be used to send information back and fourth
        bettween processes of mycroft service, voice, skill and cli
        """
        self.type = type
        self.data = data or {}
        self.context = context or {}

    def serialize(self):
        """This returns a string of the message info.

        This makes it easy to send over a websocket. This uses
        json dumps to generate the string with type, data and context

        Returns:
            str: a json string representation of the message.
        """
        return json.dumps({
            'type': self.type,
            'data': self.data,
            'context': self.context
        })

    @staticmethod
    def deserialize(value):
        """This takes a string and constructs a message object.

        This makes it easy to take strings from the websocket and create
        a message object.  This uses json loads to get the info and generate
        the message object.

        Args:
            value(str): This is the json string received from the websocket

        Returns:
            Message: message object constructed from the json string passed
            int the function.
            value(str): This is the string received from the websocket
        """
        obj = json.loads(value)
        return Message(obj.get('type') or '',
                       obj.get('data') or {},
                       obj.get('context') or {})

    def reply(self, type, data=None, context=None):
        """Construct a reply message for a given message

        This will take the same parameters as a message object but use
        the current message object as a reference.  It will copy the context
        from the existing message object and add any context passed in to
        the function.  Check for a target passed in to the function from
        the data object and add that to the context as a target.  If the
        context has a client name then that will become the target in the
        context.  The new message will then have data passed in plus the
        new context generated.

        Args:
            type (str): type of message
            data (dict): data for message
            context: intented context for new message

        Returns:
            Message: Message object to be used on the reply to the message
        """
        data = data or {}
        context = context or {}

        new_context = self.context
        for key in context:
            new_context[key] = context[key]
        if 'target' in data:
            new_context['target'] = data['target']
        elif 'client_name' in context:
            context['target'] = context['client_name']
        return Message(type, data, context=new_context)

    def response(self, data=None, context=None):
        """Construct a response message for the message

        Constructs a reply with the data and appends the expected
        ".response" to the message

        Args:
            data (dict): message data
            context (dict): message context
        Returns
            (Message) message with the type modified to match default response
        """
        response_message = self.reply(self.type, data or {}, context)
        response_message.type += '.response'
        return response_message

    def publish(self, type, data, context=None):
        """
        Copy the original context and add passed in context.  Delete
        any target in the new context. Return a new message object with
        passed in data and new context.  Type remains unchanged.

        Args:
            type (str): type of message
            data (dict): date to send with message
            context: context added to existing context

        Returns:
            Message: Message object to publish
        """
        context = context or {}
        new_context = self.context.copy()
        for key in context:
            new_context[key] = context[key]

        if 'target' in new_context:
            del new_context['target']

        return Message(type, data, context=new_context)

    def utterance_remainder(self):
        """
        For intents get the portion not consumed by Adapt.

        For example: if they say 'Turn on the family room light' and there are
        entity matches for "turn on" and "light", then it will leave behind
        " the family room " which is then normalized to "family room".

        Returns:
            str: Leftover words or None if not an utterance.
        """
        utt = normalize(self.data.get("utterance", ""))
        if utt and "__tags__" in self.data:
            for token in self.data["__tags__"]:
                # Substitute only whole words matching the token
                utt = re.sub(r'\b' + token.get("key", "") + r"\b", "", utt)
        return normalize(utt)



import os
from os.path import join, dirname, isdir, isfile, exists, abspath, splitext, basename

from pocketsphinx import pocketsphinx, Jsgf, FsgModel
from speech_recognition import AudioData, RequestError, \
    PortableNamedTemporaryFile, UnknownValueError


class PS_Recognizer(object):
    def __init__(self, language="en-US", language_directory=None,
                 acoustic_parameters_directory=None,
                 language_model_file=None, phoneme_dictionary_file=None):
        super(PS_Recognizer, self).__init__()
        language = language.lower()
        language_directory = language_directory or join(
            dirname(dirname(__file__)), "recognizer/model",
            language)
        if not isdir(language_directory):
            raise RequestError(
                "missing PocketSphinx language data directory: \"{}\"".format(
                    language_directory))

        acoustic_parameters_directory = \
            acoustic_parameters_directory or \
            join(language_directory, "hmm")
        if not isdir(acoustic_parameters_directory):
            raise RequestError(
                "missing PocketSphinx language model parameters directory: "
                "\"{}\"".format(acoustic_parameters_directory))

        language_model_file = language_model_file or join(
            language_directory, language + ".lm")
        if not isfile(language_model_file):
            language_model_file += ".bin"
            if not isfile(language_model_file):
                raise RequestError(
                    "missing PocketSphinx language model file: \"{}\"".format(
                        language_model_file))

        phoneme_dictionary_file = phoneme_dictionary_file or join(
            language_directory, language + ".dict")
        if not isfile(phoneme_dictionary_file):
            raise RequestError(
                "missing PocketSphinx phoneme dictionary file: \"{}\"".format(
                    phoneme_dictionary_file))

        # create decoder object
        config = pocketsphinx.Decoder.default_config()
        config.set_string("-hmm",
                          acoustic_parameters_directory)
        config.set_string("-lm", language_model_file)
        config.set_string("-dict", phoneme_dictionary_file)
        config.set_string("-logfn",
                          os.devnull)
        self.decoder = pocketsphinx.Decoder(config)
        self.lang = language

    def recognize(self, audio_data, keyword_entries=None, grammar=None):
        language = self.lang
        assert isinstance(audio_data,
                          AudioData), "``audio_data`` must be audio data"
        assert isinstance(language, str), "``language`` must be a string"
        assert keyword_entries is None or all(
            isinstance(keyword,
                       (type(""), type(u""))) and 0 <= sensitivity <= 1
            for keyword, sensitivity in
            keyword_entries), "``keyword_entries`` must be ``None`` or" \
                              " a list of pairs of strings and " \
                              "numbers between 0 and 1"

        # obtain audio data
        raw_data = audio_data.get_raw_data(convert_rate=16000,
                                           convert_width=2)
        # obtain recognition results
        if keyword_entries is not None:  # explicitly specified set of keywords
            with PortableNamedTemporaryFile("w") as f:
                # generate a keywords file
                f.writelines(
                    "{} /1e{}/\n".format(keyword, 100 * sensitivity - 110)
                    for keyword, sensitivity in keyword_entries)
                f.flush()

                # perform the speech recognition with the keywords file
                self.decoder.set_kws("keywords", f.name)
                self.decoder.set_search("keywords")
                self.decoder.start_utt()  # begin utterance processing
                self.decoder.process_raw(raw_data, False,
                                         True)
                self.decoder.end_utt()  # stop utterance processing
        elif grammar is not None:  # a path to a FSG or JSGF grammar

            if not exists(grammar):
                raise ValueError(
                    "Grammar '{0}' does not exist.".format(grammar))
            grammar_path = abspath(dirname(grammar))
            grammar_name = splitext(basename(grammar))[0]
            fsg_path = "{0}/{1}.fsg".format(grammar_path, grammar_name)
            if not exists(
                    fsg_path):  # create FSG grammar if not available
                jsgf = Jsgf(grammar)
                rule = jsgf.get_rule("{0}.{0}".format(grammar_name))
                fsg = jsgf.build_fsg(rule, self.decoder.get_logmath(), 7.5)
                fsg.writefile(fsg_path)
            else:
                fsg = FsgModel(fsg_path, self.decoder.get_logmath(), 7.5)
            self.decoder.set_fsg(grammar_name, fsg)
            self.decoder.set_search(grammar_name)
            self.decoder.start_utt()
            self.decoder.process_raw(raw_data, False,
                                     True)
            self.decoder.end_utt()  # stop utterance processing
        else:  # no keywords, perform freeform recognition
            self.decoder.start_utt()  # begin utterance processing
            self.decoder.process_raw(raw_data, False,
                                     True)
            self.decoder.end_utt()  # stop utterance processing

        # return results
        hypothesis = self.decoder.hyp()

        if hypothesis is not None:
            return hypothesis.hypstr
        raise UnknownValueError()  # no transcriptions available



# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import re
from abc import ABCMeta, abstractmethod

from requests import post
from speech_recognition import Recognizer
from jarbas_hive_mind.settings import STT_CONFIG


class STT(object):
    __metaclass__ = ABCMeta

    def __init__(self):
        config_core = STT_CONFIG
        self.lang = str(self.init_language(config_core))
        config_stt = config_core.get("stt", {})
        self.config = config_stt.get(config_stt.get("module"), {})
        self.credential = self.config.get("credential", {})
        self.recognizer = Recognizer()

    @staticmethod
    def init_language(config_core):
        lang = config_core.get("lang", "en-US")
        langs = lang.split("-")
        if len(langs) == 2:
            return langs[0].lower() + "-" + langs[1].upper()
        return lang

    @abstractmethod
    def execute(self, audio, language=None):
        pass


class TokenSTT(STT):
    __metaclass__ = ABCMeta

    def __init__(self):
        super(TokenSTT, self).__init__()
        # do not string cast, if token is none uses default test key
        # free google instead of no google
        self.token = self.credential.get("token")


class GoogleJsonSTT(STT):
    __metaclass__ = ABCMeta

    def __init__(self):
        super(GoogleJsonSTT, self).__init__()
        self.json_credentials = json.dumps(self.credential.get("json"))


class BasicSTT(STT):
    __metaclass__ = ABCMeta

    def __init__(self):
        super(BasicSTT, self).__init__()
        self.username = str(self.credential.get("username"))
        self.password = str(self.credential.get("password"))


class KeySTT(STT):
    __metaclass__ = ABCMeta

    def __init__(self):
        super(KeySTT, self).__init__()
        self.id = str(self.credential.get("client_id"))
        self.key = str(self.credential.get("client_key"))


class GoogleSTT(TokenSTT):
    def __init__(self):
        super(GoogleSTT, self).__init__()
        if self.token == "None":
            self.token = None

    def execute(self, audio, language=None):
        self.lang = language or self.lang
        return self.recognizer.recognize_google(audio, self.token, self.lang)


class GoogleCloudSTT(GoogleJsonSTT):
    def __init__(self):
        super(GoogleCloudSTT, self).__init__()

    def execute(self, audio, language=None):
        self.lang = language or self.lang
        return self.recognizer.recognize_google_cloud(audio,
                                                      self.json_credentials,
                                                      self.lang)


class WITSTT(TokenSTT):
    def __init__(self):
        super(WITSTT, self).__init__()

    def execute(self, audio, language=None):
        print("WITSTT language should be configured at wit.ai settings.")
        return self.recognizer.recognize_wit(audio, self.token)


class IBMSTT(BasicSTT):
    def __init__(self):
        super(IBMSTT, self).__init__()

    def execute(self, audio, language=None):
        self.lang = language or self.lang
        return self.recognizer.recognize_ibm(audio, self.username,
                                             self.password, self.lang)


class DeepSpeechServerSTT(STT):
    """
        STT interface for the deepspeech-server:
        https://github.com/MainRo/deepspeech-server
        use this if you want to host DeepSpeech yourself
    """
    def __init__(self):
        super(DeepSpeechServerSTT, self).__init__()

    def execute(self, audio, language=None):
        language = language or self.lang
        if not language.startswith("en"):
            raise ValueError("Deepspeech is currently english only")
        response = post(self.config.get("uri"), data=audio.get_wav_data())
        return response.text


class KaldiSTT(STT):
    def __init__(self):
        super(KaldiSTT, self).__init__()

    def execute(self, audio, language=None):
        language = language or self.lang
        response = post(self.config.get("uri"), data=audio.get_wav_data())
        return self.get_response(response)

    def get_response(self, response):
        try:
            hypotheses = response.json()["hypotheses"]
            return re.sub(r'\s*\[noise\]\s*', '', hypotheses[0]["utterance"])
        except:
            return None


class PocketSphinxSTT(BasicSTT):
    def __init__(self, lang="en-us", config=None):
        super(PocketSphinxSTT, self).__init__()
        from jarbas_hive_mind.terminals.speech.stt.pocketsphinx_stt import PS_Recognizer
        self.recognizer = PS_Recognizer(self.lang)

    def execute(self, audio, language=None):
        language = language or self.lang
        if language != self.lang:
            print("Changing decoder language")
            from jarbas_hive_mind.terminals.speech.stt.pocketsphinx_stt import PS_Recognizer
            self.lang = language
            self.recognizer = PS_Recognizer(self.lang)

        return self.recognizer.recognize(audio)


class BingSTT(TokenSTT):
    def __init__(self):
        super(BingSTT, self).__init__()

    def execute(self, audio, language=None):
        self.lang = language or self.lang
        return self.recognizer.recognize_bing(audio, self.token,
                                              self.lang)


class HoundifySTT(KeySTT):
    def __init__(self):
        super(HoundifySTT, self).__init__()

    def execute(self, audio, language=None):
        self.lang = language or self.lang
        return self.recognizer.recognize_houndify(audio, self.id, self.key)


class STTFactory(object):
    CLASSES = {
        "google": GoogleSTT,
        "google_cloud": GoogleCloudSTT,
        "wit": WITSTT,
        "ibm": IBMSTT,
        "kaldi": KaldiSTT,
        "pocketsphinx": PocketSphinxSTT,
        "houndify": HoundifySTT,
        "bing": BingSTT,
        "deepspeech_server": DeepSpeechServerSTT
    }

    @staticmethod
    def create(config=None):
        config = config or STT_CONFIG.get("stt", {})
        module = config.get("module", "google")
        clazz = STTFactory.CLASSES.get(module)
        return clazz()



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import defaultdict
from multiprocessing.pool import ThreadPool

from pyee import EventEmitter


class ThreadedEventEmitter(EventEmitter):
    """ Event Emitter using the threadpool to run event functions in
        separate threads.
    """

    def __init__(self, threads=10):
        super().__init__()
        self.pool = ThreadPool(threads)
        self.wrappers = defaultdict(list)

    def on(self, event, f=None):
        """ Wrap on with a threaded launcher. """

        def wrapped(*args, **kwargs):
            return self.pool.apply_async(f, args, kwargs)

        w = super().on(event, wrapped)
        # Store mapping from function to wrapped function
        self.wrappers[event].append((f, wrapped))
        return w

    def once(self, event, f=None):
        """ Wrap once with a threaded launcher. """

        def wrapped(*args, **kwargs):
            return self.pool.apply_async(f, args, kwargs)

        wrapped = super().once(event, wrapped)
        self.wrappers[event].append((f, wrapped))
        return wrapped

    def remove_listener(self, event_name, func):
        """ Wrap the remove to translate from function to wrapped
            function.
        """
        for w in self.wrappers[event_name]:
            if w[0] == func:
                self.wrappers[event_name].remove(w)
                return super().remove_listener(event_name, w[1])
        # if no wrapper exists try removing the function
        return super().remove_listener(event_name, func)



from jarbas_hive_mind.terminals.https import JarbasMicroServicesAPI, logger


def connect_to_hivenode(key="test_key",
                        url="https://0.0.0.0:6712/"):

    ap = JarbasMicroServicesAPI(key, url)
    while True:
        line = input("Input: ")
        res = ap.ask_mycroft(line.lower())
        logger.info("Jarbas: " + res.get("data", {})
                    .get("utterance", "does not compute"))


if __name__ == "__main__":
    # TODO argparse
    connect_to_hivenode()




# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import audioop
import collections
import datetime
import logging
import sys
from hashlib import md5
from os import mkdir
from os.path import isdir, join
from subprocess import check_output, Popen
from tempfile import gettempdir
from threading import Lock
from time import sleep, time as get_time

import pyaudio
import requests
import speech_recognition
from speech_recognition import (
    Microphone,
    AudioSource,
    AudioData
)

from jarbas_hive_mind.terminals.speech.signal import *

logger = logging.getLogger("mic")
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel("INFO")

conf = {
    "listener": {
        "sample_rate": 16000,
        "channels": 1,
        "record_wake_words": False,
        "record_utterances": False,
        "phoneme_duration": 120,
        "multiplier": 1.0,
        "energy_ratio": 1.5,
        "wake_word": "hey mycroft",
        "stand_up_word": "wake up"
      },
    "hotwords": {
        "hey mycroft": {
            "module": "pocketsphinx",
            "phonemes": "HH EY . M AY K R AO F T",
            "threshold": 1e-90,
            "lang": "en-us"
        },
        "thank you": {
            "module": "pocketsphinx",
            "phonemes": "TH AE NG K . Y UW .",
            "threshold": 1e-1,
            "listen": False,
            "utterance": "thank you",
            "active": True,
            "sound": "",
            "lang": "en-us"
        },
        "wake up": {
            "module": "pocketsphinx",
            "phonemes": "W EY K . AH P",
            "threshold": 1e-20,
            "lang": "en-us"
        }
    },
    "sounds": {
        "start_listening": "start_listening.wav"
      },
}


def get_http(uri):
    return uri.replace("https://", "http://")


def play_wav(uri):
    play_cmd = "paplay %1 --stream-name=mycroft-voice"
    play_wav_cmd = str(play_cmd).split(" ")
    for index, cmd in enumerate(play_wav_cmd):
        if cmd == "%1":
            play_wav_cmd[index] = (get_http(uri))
    return Popen(play_wav_cmd)


class MutableStream(object):
    def __init__(self, wrapped_stream, format, muted=False):
        assert wrapped_stream is not None
        self.wrapped_stream = wrapped_stream
        self.muted = muted

        self.SAMPLE_WIDTH = pyaudio.get_sample_size(format)
        self.muted_buffer = b''.join([b'\x00' * self.SAMPLE_WIDTH])

    def mute(self):
        self.muted = True

    def unmute(self):
        self.muted = False

    def read(self, size, of_exc=False):
        """
            Read data from stream.

            Arguments:
                size (int): Number of bytes to read
                of_exc (bool): flag determining if the audio producer thread
                               should throw IOError at overflows.

            Returns:
                Data read from device
        """
        frames = collections.deque()
        remaining = size
        while remaining > 0:
            to_read = min(self.wrapped_stream.get_read_available(), remaining)
            if to_read == 0:
                sleep(.01)
                continue
            result = self.wrapped_stream.read(to_read,
                                              exception_on_overflow=of_exc)
            frames.append(result)
            remaining -= to_read

        if self.muted:
            return self.muted_buffer
        input_latency = self.wrapped_stream.get_input_latency()
        if input_latency > 0.2:
            logger.warning("[WARNING] High input latency: %f" % input_latency)
        audio = b"".join(list(frames))
        return audio

    def close(self):
        self.wrapped_stream.close()
        self.wrapped_stream = None

    def is_stopped(self):
        return self.wrapped_stream.is_stopped()

    def stop_stream(self):
        return self.wrapped_stream.stop_stream()


class MutableMicrophone(Microphone):
    def __init__(self, device_index=None, sample_rate=16000, chunk_size=1024,
                 mute=False):
        Microphone.__init__(
            self, device_index=device_index, sample_rate=sample_rate,
            chunk_size=chunk_size)
        self.muted = False
        if mute:
            self.mute()

    def __enter__(self):
        assert self.stream is None, \
            "This audio source is already inside a context manager"
        self.audio = pyaudio.PyAudio()
        self.stream = MutableStream(self.audio.open(
            input_device_index=self.device_index, channels=1,
            format=self.format, rate=self.SAMPLE_RATE,
            frames_per_buffer=self.CHUNK,
            input=True,  # stream is an input stream
        ), self.format, self.muted)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if not self.stream.is_stopped():
            self.stream.stop_stream()
        self.stream.close()
        self.stream = None
        self.audio.terminate()

    def mute(self):
        self.muted = True
        if self.stream:
            self.stream.mute()

    def unmute(self):
        self.muted = False
        if self.stream:
            self.stream.unmute()

    def is_muted(self):
        return self.muted


def get_silence(num_bytes):
    return b'\0' * num_bytes


class ResponsiveRecognizer(speech_recognition.Recognizer):
    # Padding of silence when feeding to pocketsphinx
    SILENCE_SEC = 0.01

    # The minimum seconds of noise before a
    # phrase can be considered complete
    MIN_LOUD_SEC_PER_PHRASE = 0.5

    # The minimum seconds of silence required at the end
    # before a phrase will be considered complete
    MIN_SILENCE_AT_END = 0.25

    # The maximum seconds a phrase can be recorded,
    # provided there is noise the entire time
    RECORDING_TIMEOUT = 10.0

    # The maximum time it will continue to record silence
    # when not enough noise has been detected
    RECORDING_TIMEOUT_WITH_SILENCE = 3.0

    # Time between pocketsphinx checks for the wake word
    SEC_BETWEEN_WW_CHECKS = 0.2

    def __init__(self, wake_word_recognizer, hot_word_engines=None):

        self.config = conf
        listener_config = self.config.get('listener')
        self.upload_config = {"enable": False}
        # self.upload_config = listener_config.get('wake_word_upload')
        self.wake_word_name = wake_word_recognizer.key_phrase

        self.overflow_exc = listener_config.get('overflow_exception', False)

        speech_recognition.Recognizer.__init__(self)
        self.wake_word_recognizer = wake_word_recognizer
        self.audio = pyaudio.PyAudio()
        self.multiplier = listener_config.get('multiplier')
        self.energy_ratio = listener_config.get('energy_ratio')
        # check the config for the flag to save wake words.

        self.save_utterances = listener_config.get('record_utterances', False)
        self.save_wake_words = listener_config.get('record_wake_words')  # \
        # or self.upload_config['enable'] or self.config['opt_in']
        self.upload_lock = Lock()
        self.save_wake_words_dir = join(gettempdir(), 'mycroft_wake_words')
        self.filenames_to_upload = []
        self.mic_level_file = os.path.join(get_ipc_directory(), "mic_level")
        self._stop_signaled = False
        self.hot_word_engines = hot_word_engines or {}

        # The maximum audio in seconds to keep for transcribing a phrase
        # The wake word must fit in this time
        num_phonemes = wake_word_recognizer.num_phonemes
        len_phoneme = listener_config.get('phoneme_duration', 120) / 1000.0
        self.TEST_WW_SEC = num_phonemes * len_phoneme
        self.SAVED_WW_SEC = 10 if self.save_wake_words else self.TEST_WW_SEC

        try:
            self.account_id = "666"#DeviceApi().get()['user']['uuid']
        except (requests.HTTPError, requests.ConnectionError, AttributeError):
            self.account_id = '0'

    def record_sound_chunk(self, source):
        return source.stream.read(source.CHUNK, self.overflow_exc)

    @staticmethod
    def calc_energy(sound_chunk, sample_width):
        return audioop.rms(sound_chunk, sample_width)

    def _record_phrase(self, source, sec_per_buffer):
        """Record an entire spoken phrase.

        Essentially, this code waits for a period of silence and then returns
        the audio.  If silence isn't detected, it will terminate and return
        a buffer of RECORDING_TIMEOUT duration.

        Args:
            source (AudioSource):  Source producing the audio chunks
            sec_per_buffer (float):  Fractional number of seconds in each chunk

        Returns:
            bytearray: complete audio buffer recorded, including any
                       silence at the end of the user's utterance
        """

        num_loud_chunks = 0
        noise = 0

        max_noise = 25
        min_noise = 0

        silence_duration = 0

        def increase_noise(level):
            if level < max_noise:
                return level + 200 * sec_per_buffer
            return level

        def decrease_noise(level):
            if level > min_noise:
                return level - 100 * sec_per_buffer
            return level

        # Smallest number of loud chunks required to return
        min_loud_chunks = int(self.MIN_LOUD_SEC_PER_PHRASE / sec_per_buffer)

        # Maximum number of chunks to record before timing out
        max_chunks = int(self.RECORDING_TIMEOUT / sec_per_buffer)
        num_chunks = 0

        # Will return if exceeded this even if there's not enough loud chunks
        max_chunks_of_silence = int(self.RECORDING_TIMEOUT_WITH_SILENCE /
                                    sec_per_buffer)

        # bytearray to store audio in
        byte_data = get_silence(source.SAMPLE_WIDTH)

        phrase_complete = False
        while num_chunks < max_chunks and not phrase_complete:
            chunk = self.record_sound_chunk(source)
            byte_data += chunk
            num_chunks += 1

            energy = self.calc_energy(chunk, source.SAMPLE_WIDTH)
            test_threshold = self.energy_threshold * self.multiplier
            is_loud = energy > test_threshold
            if is_loud:
                noise = increase_noise(noise)
                num_loud_chunks += 1
            else:
                noise = decrease_noise(noise)
                self._adjust_threshold(energy, sec_per_buffer)

            if num_chunks % 10 == 0:
                with open(self.mic_level_file, 'w') as f:
                    f.write("Energy:  cur=" + str(energy) + " thresh=" +
                            str(self.energy_threshold))
                f.close()

            was_loud_enough = num_loud_chunks > min_loud_chunks

            quiet_enough = noise <= min_noise
            if quiet_enough:
                silence_duration += sec_per_buffer
                if silence_duration < self.MIN_SILENCE_AT_END:
                    quiet_enough = False  # gotta be silent for min of 1/4 sec
            else:
                silence_duration = 0
            recorded_too_much_silence = num_chunks > max_chunks_of_silence
            if quiet_enough and (was_loud_enough or recorded_too_much_silence):
                phrase_complete = True

        return byte_data

    @staticmethod
    def sec_to_bytes(sec, source):
        return int(sec * source.SAMPLE_RATE) * source.SAMPLE_WIDTH

    def _skip_wake_word(self):
        # Check if told programatically to skip the wake word, like
        # when we are in a dialog with the user.
        if check_for_signal('startListening'):
            return True

        # Pressing the Mark 1 button can start recording (unless
        # it is being used to mean 'stop' instead)
        if check_for_signal('buttonPress', 1):
            # give other processes time to consume this signal if
            # it was meant to be a 'stop'
            sleep(0.25)
            if check_for_signal('buttonPress'):
                # Signal is still here, assume it was intended to
                # begin recording
                logger.info("[INFO] Button Pressed, wakeword not needed")
                return True

        return False

    def stop(self):
        """
            Signal stop and exit waiting state.
        """
        self._stop_signaled = True

    def _wait_until_wake_word(self, source, sec_per_buffer, emitter):
        """Listen continuously on source until a wake word is spoken

        Args:
            source (AudioSource):  Source producing the audio chunks
            sec_per_buffer (float):  Fractional number of seconds in each chunk
        """
        num_silent_bytes = int(self.SILENCE_SEC * source.SAMPLE_RATE *
                               source.SAMPLE_WIDTH)

        silence = get_silence(num_silent_bytes)

        # bytearray to store audio in
        byte_data = silence

        buffers_per_check = self.SEC_BETWEEN_WW_CHECKS / sec_per_buffer
        buffers_since_check = 0.0

        # Max bytes for byte_data before audio is removed from the front
        max_size = self.sec_to_bytes(self.SAVED_WW_SEC, source)
        test_size = self.sec_to_bytes(self.TEST_WW_SEC, source)

        said_wake_word = False

        # Rolling buffer to track the audio energy (loudness) heard on
        # the source recently.  An average audio energy is maintained
        # based on these levels.
        energies = []
        idx_energy = 0
        avg_energy = 0.0
        energy_avg_samples = int(5 / sec_per_buffer)  # avg over last 5 secs

        ww_module = self.wake_word_recognizer.__class__.__name__
        if ww_module == 'PreciseHotword':
            _, model_path = self.wake_word_recognizer.get_model_info()
            model_hash = check_output(['md5sum', model_path]).split()[0]
        else:
            model_hash = '0'
        counter = 0

        while not said_wake_word and not self._stop_signaled:
            if self._skip_wake_word():
                break
            chunk = self.record_sound_chunk(source)

            energy = self.calc_energy(chunk, source.SAMPLE_WIDTH)
            if energy < self.energy_threshold * self.multiplier:
                self._adjust_threshold(energy, sec_per_buffer)

            if len(energies) < energy_avg_samples:
                # build the average
                energies.append(energy)
                avg_energy += float(energy) / energy_avg_samples
            else:
                # maintain the running average and rolling buffer
                avg_energy -= float(energies[idx_energy]) / energy_avg_samples
                avg_energy += float(energy) / energy_avg_samples
                energies[idx_energy] = energy
                idx_energy = (idx_energy + 1) % energy_avg_samples

                # maintain the threshold using average
                if energy < avg_energy * 1.5:
                    if energy > self.energy_threshold:
                        # bump the threshold to just above this value
                        self.energy_threshold = energy * 1.2

            # Periodically output energy level stats.  This can be used to
            # visualize the microphone input, e.g. a needle on a meter.
            if counter % 3:
                with open(self.mic_level_file, 'w') as f:
                    f.write("Energy:  cur=" + str(energy) + " thresh=" +
                            str(self.energy_threshold))
                f.close()
            counter += 1

            # At first, the buffer is empty and must fill up.  After that
            # just drop the first chunk bytes to keep it the same size.
            needs_to_grow = len(byte_data) < max_size
            if needs_to_grow:
                byte_data += chunk
            else:  # Remove beginning of audio and add new chunk to end
                byte_data = byte_data[len(chunk):] + chunk

            buffers_since_check += 1.0
            self.wake_word_recognizer.update(chunk)
            if buffers_since_check > buffers_per_check:
                buffers_since_check -= buffers_per_check
                chopped = byte_data[-test_size:] \
                    if test_size < len(byte_data) else byte_data
                audio_data = chopped + silence
                said_wake_word = \
                    self.wake_word_recognizer.found_wake_word(audio_data)
                if said_wake_word:
                    payload = {
                        'hotword': self.wake_word_recognizer.key_phrase,
                        'start_listening': True,
                        'sound': self.config.get('sounds', {}).get('start_listening'),
                        "engine": self.wake_word_recognizer.module
                    }
                    emitter.emit("recognizer_loop:hotword", payload)
                    # If enabled, play a wave file with a short sound
                    # to audibly indicate recording has begun.
                    if self.config.get('confirm_listening'):
                        file = self.config.get('sounds', {}).get('start_listening')
                        if file:
                            play_wav(file)
                    # if a wake word is success full then record audio in temp
                    # file.
                    if self.save_wake_words:
                        audio = self._create_audio_data(byte_data, source)

                        if not isdir(self.save_wake_words_dir):
                            mkdir(self.save_wake_words_dir)
                        dr = self.save_wake_words_dir

                        components = [
                            self.wake_word_name.replace(' ', '-'),
                            md5(ww_module.encode('utf-8')).hexdigest(),
                            str(int(1000 * get_time())),
                            model_hash
                        ]

                        fn = join(dr, '.'.join(components) + '.wav')
                        with open(fn, 'wb') as f:
                            f.write(audio.get_wav_data())

                else:
                    said_wake_word, said_hot_word = self.check_for_hotwords(
                        audio_data, emitter)
                    if said_hot_word:
                        # reset bytearray to store audio in, else many
                        # serial detections
                        byte_data = silence

    def check_for_hotwords(self, audio_data, emitter):
        # check hot word
        for hotword in self.hot_word_engines:
            engine, ding, utterance, listen, engine_type = \
                    self.hot_word_engines[hotword]
            found = engine.found_wake_word(audio_data)
            if found:
                logger.info("[INFO] Hot Word: " + hotword)
                # If enabled, play a wave file with a short sound to audibly
                # indicate hotword was detected.
                if ding:
                    try:
                        file = ding
                        if file:
                            play_wav(file)
                    except Exception as e:
                        logger.exception(e)
                # Hot Word succeeded
                payload = {
                    'hotword': hotword,
                    'start_listening': listen,
                    'sound': ding,
                    "engine": engine_type
                }
                emitter.emit("recognizer_loop:hotword", payload)
                if utterance:
                    # send the transcribed word on for processing
                    payload = {
                        'utterances': [utterance]
                    }
                    emitter.emit("recognizer_loop:utterance", payload)
                    return False, True
                if listen:
                    # start listening
                    return True, True
                return False, True
        return False, False

    @staticmethod
    def _create_audio_data(raw_data, source):
        """
        Constructs an AudioData instance with the same parameters
        as the source and the specified frame_data
        """
        return AudioData(raw_data, source.SAMPLE_RATE, source.SAMPLE_WIDTH)

    def listen(self, source, emitter):
        """Listens for chunks of audio that Mycroft should perform STT on.

        This will listen continuously for a wake-up-word, then return the
        audio chunk containing the spoken phrase that comes immediately
        afterwards.

        Args:
            source (AudioSource):  Source producing the audio chunks
            emitter (EventEmitter): Emitter for notifications of when recording
                                    begins and ends.

        Returns:
            AudioData: audio with the user's utterance, minus the wake-up-word
        """
        assert isinstance(source, AudioSource), "Source must be an AudioSource"

        #        bytes_per_sec = source.SAMPLE_RATE * source.SAMPLE_WIDTH
        sec_per_buffer = float(source.CHUNK) / source.SAMPLE_RATE

        # Every time a new 'listen()' request begins, reset the threshold
        # used for silence detection.  This is as good of a reset point as
        # any, as we expect the user and Mycroft to not be talking.
        # NOTE: adjust_for_ambient_noise() doc claims it will stop early if
        #       speech is detected, but there is no code to actually do that.
        self.adjust_for_ambient_noise(source, 1.0)

        logger.info("[INFO] Waiting for wake word...")
        self._wait_until_wake_word(source, sec_per_buffer, emitter)
        if self._stop_signaled:
            return

        emitter.emit("recognizer_loop:record_begin")

        frame_data = self._record_phrase(source, sec_per_buffer)
        audio_data = self._create_audio_data(frame_data, source)
        emitter.emit("recognizer_loop:record_end")
        if self.save_utterances:
            logger.info("[INFO] Saving utterance")
            stamp = str(datetime.datetime.now())
            filename = "/tmp/mycroft_utterance%s.wav" % stamp
            with open(filename, 'wb') as filea:
                filea.write(audio_data.get_wav_data())

        return audio_data

    def _adjust_threshold(self, energy, seconds_per_buffer):
        if self.dynamic_energy_threshold and energy > 0:
            # account for different chunk sizes and rates
            damping = (
                self.dynamic_energy_adjustment_damping ** seconds_per_buffer)
            target_energy = energy * self.energy_ratio
            self.energy_threshold = (
                self.energy_threshold * damping +
                target_energy * (1 - damping))



import base64
from jarbas_hive_mind.bridges.facebook import platform, logger, JarbasFacebookBridge
from twisted.internet import reactor, ssl


def connect_to_facebook(mail, password, host="127.0.0.1", port=5678,
                        name="Jarbas Facebook Bridge", api="fb_key",
                        useragent=platform):
    authorization = bytes(name + ":" + api, encoding="utf-8")
    usernamePasswordDecoded = authorization
    api = base64.b64encode(usernamePasswordDecoded)
    headers = {'authorization': api}
    address = u"wss://" + host + u":" + str(port)
    logger.info("[INFO] connecting to hive mind at " + address)
    bridge = JarbasFacebookBridge(mail=mail,
                                  password=password,
                                  name=name, headers=headers,
                                  useragent=useragent)
    contextFactory = ssl.ClientContextFactory()
    reactor.connectSSL(host, port, bridge, contextFactory)
    reactor.run()


if __name__ == '__main__':
    # TODO arg parse
    mail = "xx@mail.com"
    pwd = "NotThis"
    connect_to_facebook(mail, pwd)



import base64
from jarbas_hive_mind.bridges.hackchat import platform, logger, JarbasHackChatBridge
from twisted.internet import reactor, ssl


def connect_to_hackchat(channel, username="Jarbas_BOT", host="127.0.0.1",
                        port=5678, name="Jarbas HackChat Bridge",
                        api="hackchat_key", useragent=platform):
    authorization = bytes(name + ":" + api, encoding="utf-8")
    usernamePasswordDecoded = authorization
    api = base64.b64encode(usernamePasswordDecoded)
    headers = {'authorization': api}
    address = u"wss://" + host + u":" + str(port)
    logger.info("[INFO] connected to hivemind: " + address)
    bridge = JarbasHackChatBridge(channel=channel, username=username,
                                  headers=headers, useragent=useragent)
    contextFactory = ssl.ClientContextFactory()
    reactor.connectSSL(host, port, bridge, contextFactory)
    reactor.run()


if __name__ == '__main__':
    # TODO arg parse
    connect_to_hackchat("JarbasAI_bot")



import base64
from twisted.internet import reactor, ssl
from jarbas_hive_mind.bridges.twitter import platform, logger, JarbasTwitterBridge


def connect_to_twitter(consumer_key, consumer_secret, access_token,
                       access_token_secret, users, languages=None,
                       host="127.0.0.1", port=5678,
                       name="Jarbas Twitter Bridge", api="twitter_key",
                       useragent=platform):
    authorization = bytes(name + ":" + api, encoding="utf-8")
    usernamePasswordDecoded = authorization
    api = base64.b64encode(usernamePasswordDecoded)
    headers = {'authorization': api}
    address = u"wss://" + host + u":" + str(port)
    logger.info("[INFO] connecting to hive mind at " + address)
    bridge = JarbasTwitterBridge(consumer_key, consumer_secret, access_token,
                                 access_token_secret, users, languages,
                                 name=name, headers=headers,
                                 useragent=useragent)
    contextFactory = ssl.ClientContextFactory()
    reactor.connectSSL(host, port, bridge, contextFactory)
    reactor.run()


if __name__ == '__main__':
    # TODO arg parse
    CONSUMER_KEY = 'xx'
    CONSUMER_SECRET = 'Cxxx'
    ACCESS_TOKEN = 'xxxIy'
    ACCESS_TOKEN_SECRET = 'bvUxx'

    USERS = ['@jarbasAi']

    LANGUAGES = ['en']

    connect_to_twitter(CONSUMER_KEY, CONSUMER_SECRET, ACCESS_TOKEN,
                       ACCESS_TOKEN_SECRET, USERS, LANGUAGES)



# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import time
import traceback
from threading import Event

from websocket import (WebSocketApp, WebSocketConnectionClosedException,
                       WebSocketException)

from jarbas_hive_mind.settings import MYCROFT_WEBSOCKET_CONFIG
from jarbas_hive_mind.utils import validate_param, create_echo_function
from jarbas_hive_mind.utils.log import LOG
from jarbas_hive_mind.utils.messagebus.message import Message
from jarbas_hive_mind.utils.messagebus.threaded_event_emitter import \
    ThreadedEventEmitter


class WebsocketClient:
    def __init__(self, host=None, port=None, route=None, ssl=None,
                 config=MYCROFT_WEBSOCKET_CONFIG):
        host = host or config.get("host")
        port = port or config.get("port")
        route = route or config.get("route")
        ssl = ssl or config.get("ssl")
        validate_param(host, "websocket.host")
        validate_param(port, "websocket.port")
        validate_param(route, "websocket.route")

        self.url = WebsocketClient.build_url(host, port, route, ssl)
        self.emitter = ThreadedEventEmitter()
        self.client = self.create_client()
        self.retry = 5
        self.connected_event = Event()
        self.started_running = False

    @staticmethod
    def build_url(host, port, route, ssl):
        scheme = "wss" if ssl else "ws"
        return scheme + "://" + host + ":" + str(port) + route

    def create_client(self):
        return WebSocketApp(self.url,
                            on_open=self.on_open, on_close=self.on_close,
                            on_error=self.on_error, on_message=self.on_message)

    def on_open(self):
        LOG.info("Connected")
        self.connected_event.set()
        self.emitter.emit("open")
        # Restore reconnect timer to 5 seconds on sucessful connect
        self.retry = 5

    def on_close(self):
        self.emitter.emit("close")

    def on_error(self, error):
        """ On error start trying to reconnect to the websocket. """
        if isinstance(error, WebSocketConnectionClosedException):
            LOG.warning('Could not send message because connection has closed')
        else:
            LOG.exception('=== ' + repr(error) + ' ===')

        try:
            self.emitter.emit('error', error)
            if self.client.keep_running:
                self.client.close()
        except Exception as e:
            LOG.error('Exception closing websocket: ' + repr(e))

        LOG.warning("WS Client will reconnect in %d seconds." % self.retry)
        time.sleep(self.retry)
        self.retry = min(self.retry * 2, 60)
        try:
            self.emitter.emit('reconnecting')
            self.client = self.create_client()
            self.run_forever()
        except WebSocketException:
            pass

    def on_message(self, message):
        parsed_message = Message.deserialize(message)
        self.emitter.emit(parsed_message.type, parsed_message)
        self.emitter.emit('message', message)

    def emit(self, message):
        if not self.connected_event.wait(10):
            if not self.started_running:
                raise ValueError('You must execute run_forever() '
                                 'before emitting messages')
            self.connected_event.wait()

        try:
            if hasattr(message, 'serialize'):
                self.client.send(message.serialize())
            else:
                self.client.send(json.dumps(message.__dict__))
        except WebSocketConnectionClosedException:
            LOG.warning('Could not send {} message because connection '
                        'has been closed'.format(message.type))

    def wait_for_response(self, message, reply_type=None, timeout=None):
        """Send a message and wait for a response.

        Args:
            message (Message): message to send
            reply_type (str): the message type of the expected reply.
                              Defaults to "<message.type>.response".
            timeout: seconds to wait before timeout, defaults to 3
        Returns:
            The received message or None if the response timed out
        """
        response = []

        def handler(message):
            """Receive response data."""
            response.append(message)

        # Setup response handler
        self.once(reply_type or message.type + '.response', handler)
        # Send request
        self.emit(message)
        # Wait for response
        start_time = time.monotonic()
        while len(response) == 0:
            time.sleep(0.2)
            if time.monotonic() - start_time > (timeout or 3.0):
                try:
                    self.remove(reply_type, handler)
                except (ValueError, KeyError):
                    # ValueError occurs on pyee 1.0.1 removing handlers
                    # registered with once.
                    # KeyError may theoretically occur if the event occurs as
                    # the handler is removed
                    pass
                return None
        return response[0]

    def on(self, event_name, func):
        self.emitter.on(event_name, func)

    def once(self, event_name, func):
        self.emitter.once(event_name, func)

    def remove(self, event_name, func):
        try:
            if event_name in self.emitter._events:
                LOG.debug("Removing found '" + str(event_name) + "'")
            else:
                LOG.debug("Not able to find '" + str(event_name) + "'")
            self.emitter.remove_listener(event_name, func)
        except ValueError:
            LOG.warning('Failed to remove event {}: {}'.format(event_name,
                                                               str(func)))
            for line in traceback.format_stack():
                LOG.warning(line.strip())

            if event_name in self.emitter._events:
                LOG.debug("Removing found '" + str(event_name) + "'")
            else:
                LOG.debug("Not able to find '" + str(event_name) + "'")
            LOG.warning("Existing events: " + str(self.emitter._events))
            for evt in self.emitter._events:
                LOG.warning("   " + str(evt))
                LOG.warning("       " + str(self.emitter._events[evt]))
            if event_name in self.emitter._events:
                LOG.debug("Removing found '" + str(event_name) + "'")
            else:
                LOG.debug("Not able to find '" + str(event_name) + "'")
            LOG.warning('----- End dump -----')

    def remove_all_listeners(self, event_name):
        '''
            Remove all listeners connected to event_name.

            Args:
                event_name: event from which to remove listeners
        '''
        if event_name is None:
            raise ValueError
        self.emitter.remove_all_listeners(event_name)

    def run_forever(self):
        self.started_running = True
        self.client.run_forever()

    def close(self):
        self.client.close()
        self.connected_event.clear()


def echo():
    ws = WebsocketClient()

    def repeat_utterance(message):
        message.type = 'speak'
        ws.emit(message)

    ws.on('message', create_echo_function(None))
    ws.on('recognizer_loop:utterance', repeat_utterance)
    ws.run_forever()


if __name__ == "__main__":
    echo()



from threading import Thread

from autobahn.twisted.websocket import WebSocketClientFactory, \
    WebSocketClientProtocol
from twisted.internet import reactor, ssl
from twisted.internet.protocol import ReconnectingClientFactory

from jarbas_hive_mind.terminals.speech.listener import RecognizerLoop
from jarbas_hive_mind.settings import LISTENER_CONFIG

import json
import sys
import logging
import base64
from responsive_voice import ResponsiveVoice

platform = "JarbasVoiceTerminalv0.1"

logger = logging.getLogger(platform)
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel("INFO")


class JarbasVoiceTerminalProtocol(WebSocketClientProtocol):

    def onConnect(self, response):
        logger.info("[INFO] Server connected: {0}".format(response.peer))
        self.factory.client = self
        self.factory.status = "connected"

    def onOpen(self):
        logger.info("[INFO] WebSocket connection open. ")
        self.loop = RecognizerLoop(self.factory.config)
        self.listen = Thread(target=self.start_listening)
        self.listen.setDaemon(True)
        self.listen.start()

    def handle_record_begin(self):
        logger.info("[INFO] Begin Recording...")

    def handle_record_end(self):
        logger.info("[INFO] End Recording...")

    def handle_awoken(self):
        """ Forward mycroft.awoken to the messagebus. """
        logger.info("[INFO] Listener is now Awake: ")

    def handle_wakeword(self, event):
        logger.info("[INFO] Wakeword Detected: " + event['utterance'])

    def handle_utterance(self, event):
        context = {'platform': platform, "source": self.peer,
                   'destinatary': "hive_mind"}
        msg = {"data": {"utterances": event['utterances'], "lang": "en-us"},
               "type": "recognizer_loop:utterance",
               "context": context}

        self.send(msg)

    def handle_unknown(self):
        logger.info("[INFO] mycroft.speech.recognition.unknown")

    def handle_hotword(self, event):
        config = self.factory.config.get("listener", {})
        ww = config.get("wake_word", "hey mycroft")
        suw = config.get("stand_up_word", "wake up")
        if event["hotword"] != ww and event["hotword"] != suw:
            logger.info("[INFO] Hotword Detected: " + event['hotword'])

    def handle_sleep(self):
        self.loop.sleep()

    def handle_wake_up(self, event):
        self.loop.awaken()

    def handle_mic_mute(self, event):
        self.loop.mute()

    def handle_mic_unmute(self, event):
        self.loop.unmute()

    def handle_audio_start(self, event):
        """
            Mute recognizer loop
        """
        self.loop.mute()

    def handle_audio_end(self, event):
        """
            Request unmute, if more sources has requested the mic to be muted
            it will remain muted.
        """
        self.loop.unmute()  # restore

    def handle_stop(self, event):
        """
            Handler for mycroft.stop, i.e. button press
        """
        self.loop.force_unmute()

    def start_listening(self):
        self.loop.on('recognizer_loop:utterance', self.handle_utterance)
        self.loop.on('recognizer_loop:record_begin', self.handle_record_begin)
        self.loop.on('recognizer_loop:awoken', self.handle_awoken)
        self.loop.on('recognizer_loop:wakeword', self.handle_wakeword)
        self.loop.on('recognizer_loop:hotword', self.handle_hotword)
        self.loop.on('recognizer_loop:record_end', self.handle_record_end)
        self.loop.run()

    def stop_listening(self):
        self.loop.remove_listener('recognizer_loop:utterance',
                                  self.handle_utterance)
        self.loop.remove_listener('recognizer_loop:record_begin',
                                  self.handle_record_begin)
        self.loop.remove_listener('recognizer_loop:awoken', self.handle_awoken)
        self.loop.remove_listener('recognizer_loop:wakeword',
                                  self.handle_wakeword)
        self.loop.remove_listener('recognizer_loop:hotword',
                                  self.handle_hotword)
        self.loop.remove_listener('recognizer_loop:record_end',
                                  self.handle_record_end)
        self.listen.join(0)

    def onMessage(self, payload, isBinary):
        if not isBinary:
            payload = payload.decode("utf-8")
            msg = json.loads(payload)
            if msg.get("type", "") == "speak":
                utterance = msg["data"]["utterance"]
                logger.info("[OUTPUT] " + utterance)
                self.factory.engine.say(utterance)
            elif msg.get("type", "") == "hive.complete_intent_failure":
                logger.error("[ERROR] complete intent failure")
        else:
            pass

    def send(self, msg):
        msg = json.dumps(msg)
        msg = bytes(msg, encoding="utf-8")
        self.sendMessage(msg, False)

    def onClose(self, wasClean, code, reason):
        logger.info(
            "[INFO] WebSocket connection closed: {0}".format(reason))
        self.stop_listening()
        self.factory.client = None
        self.factory.status = "disconnected"
        if "Internalservererror:InvalidAPIkey" in reason:
            logger.error("[ERROR] invalid user:key provided")
            utterance = "hive mind refused connection, invalid user or key " \
                        "provided"
            self.factory.engine.say(utterance)
            raise ConnectionAbortedError("invalid user:key provided")


class JarbasVoiceTerminal(WebSocketClientFactory,
                          ReconnectingClientFactory):
    protocol = JarbasVoiceTerminalProtocol

    def __init__(self, config=LISTENER_CONFIG, *args, **kwargs):
        super(JarbasVoiceTerminal, self).__init__(*args, **kwargs)
        self.status = "disconnected"
        self.client = None
        # TODO make optional
        self.engine = ResponsiveVoice(gender="female")
        self.config = config

    # websocket handlers
    def clientConnectionFailed(self, connector, reason):
        logger.info("[INFO] Client connection failed: " +
                    str(reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)

    def clientConnectionLost(self, connector, reason):
        logger.info("[INFO] Client connection lost: " +
                    str(reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)


def connect_to_hivemind(config=LISTENER_CONFIG, host="127.0.0.1",
                        port=5678, name="Jarbas Voice Terminal",
                        key="voice_key", useragent=platform):
    authorization = bytes(name + ":" + key, encoding="utf-8")
    usernamePasswordDecoded = authorization
    key = base64.b64encode(usernamePasswordDecoded)

    headers = {'authorization': key}
    address = u"wss://" + host + u":" + str(port)
    logger.info("[INFO] connecting to hive mind at " + address)
    terminal = JarbasVoiceTerminal(config=config, headers=headers,
                                   useragent=useragent)
    contextFactory = ssl.ClientContextFactory()
    reactor.connectSSL(host, port, terminal, contextFactory)
    reactor.run()


if __name__ == '__main__':
    # TODO parse args
    connect_to_hivemind()



import logging
import sys
from threading import Thread

from autobahn.twisted.websocket import WebSocketClientFactory, \
    WebSocketClientProtocol
from twisted.internet.protocol import ReconnectingClientFactory

platform = "JarbasTwitterBridgev0.1"
logger = logging.getLogger(platform)
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel("INFO")

import json

from twitter import Api


class TwitterMonitor:
    def __init__(self, consumer_key, consumer_secret, access_token,
                 access_token_secret, users=None, languages=None):
        if isinstance(users, str):
            users = [users]
        # Users to watch for should be a list. This will be joined by Twitter and the
        # data returned will be for any tweet mentioning:
        # @user1 *OR* @user2 *OR* @user3.
        self.users = users or []
        # Languages to filter tweets by is a list. This will be joined by Twitter
        # to return data mentioning tweets only in the english language.
        self.languages = languages or ['en']

        # Since we're going to be using a streaming endpoint, there is no need to worry
        # about rate limits.
        self.api = Api(consumer_key, consumer_secret,
                       access_token, access_token_secret)
        self.handle_tweet = []

    def on_tweet(self, handler):
        assert hasattr(handler, "__call__")
        self.handle_tweet += [handler]

    def run(self):
        # api.GetStreamFilter will return a generator that yields one status
        # message (i.e., Tweet) at a time as a JSON dictionary.
        for tweet in self.api.GetStreamFilter(track=self.users,
                                              languages=self.languages):
            for handler in self.handle_tweet:
                handler(tweet, self)


class JarbasTwitterBridgeProtocol(WebSocketClientProtocol):
    twitr = None
    monitor_thread = None
    CONSUMER_KEY = 'XXX'
    # CONSUMER_KEY = os.getenv("CONSUMER_KEY", None)
    CONSUMER_SECRET = 'XXX'
    # CONSUMER_SECRET = os.getenv("CONSUMER_SECRET", None)
    ACCESS_TOKEN = 'XXX'
    # ACCESS_TOKEN = os.getenv("ACCESS_TOKEN", None)
    ACCESS_TOKEN_SECRET = 'XXX'
    # ACCESS_TOKEN_SECRET = os.getenv("ACCESS_TOKEN_SECRET", None)

    USERS = ['@jarbasAi']

    LANGUAGES = ['en']

    def start_monitoring(self):
        self.twitr = TwitterMonitor(self.CONSUMER_KEY, self.CONSUMER_SECRET,
                                    self.ACCESS_TOKEN,
                                    self.ACCESS_TOKEN_SECRET,
                                    self.USERS, self.LANGUAGES)
        self.twitr.on_tweet(self.handle_tweet)
        self.twitr.run()

    def handle_tweet(self, data, twitr):
        mention = data["in_reply_to_screen_name"]
        text = data["text"]
        user = data["user"]["screen_name"]
        utterance = text.replace("@" + mention, "")
        msg = {"data": {"utterances": [utterance],
                        "lang": "en-us"},
               "type": "recognizer_loop:utterance",
               "context": {"source": self.peer,
                           "destination": "hive_mind",
                           "platform": platform,
                           "user": user}}
        msg = json.dumps(msg)
        msg = bytes(msg, encoding="utf-8")
        self.sendMessage(msg, False)

    def onConnect(self, response):
        logger.info("[INFO] Server connected: {0}".format(response.peer))
        self.factory.client = self
        self.factory.status = "connected"

        self.CONSUMER_KEY = self.factory.CONSUMER_KEY
        self.CONSUMER_SECRET = self.factory.CONSUMER_SECRET
        self.ACCESS_TOKEN = self.factory.ACCESS_TOKEN
        self.ACCESS_TOKEN_SECRET = self.factory.ACCESS_TOKEN_SECRET
        self.USERS = self.factory.USERS
        self.LANGUAGES = self.factory.LANGUAGES

    def onOpen(self):
        logger.info("[INFO] WebSocket connection open. ")
        self.monitor_thread = Thread(target=self.start_monitoring)
        self.monitor_thread.setDaemon(True)
        self.monitor_thread.start()

    def onMessage(self, payload, isBinary):
        if not isBinary:
            payload = payload.decode("utf-8")
            msg = json.loads(payload)
            utterance = ""
            if msg.get("type", "") == "speak":
                utterance = msg["data"]["utterance"]
            elif msg.get("type", "") == "server.complete_intent_failure":
                utterance = "does not compute"

            if utterance:
                user = msg["context"]["user"]
                self.twitr.api.PostUpdate("@" + user + " " + utterance)
        else:
            pass

    def onClose(self, wasClean, code, reason):
        logger.info("[INFO] WebSocket connection closed: {0}".format(reason))
        self.factory.client = None
        self.factory.status = "disconnected"


class JarbasTwitterBridge(WebSocketClientFactory,
                          ReconnectingClientFactory):
    protocol = JarbasTwitterBridgeProtocol
    CONSUMER_KEY = 'XXX'
    # CONSUMER_KEY = os.getenv("CONSUMER_KEY", None)
    CONSUMER_SECRET = 'XXX'
    # CONSUMER_SECRET = os.getenv("CONSUMER_SECRET", None)
    ACCESS_TOKEN = 'XXX'
    # ACCESS_TOKEN = os.getenv("ACCESS_TOKEN", None)
    ACCESS_TOKEN_SECRET = 'XXX'
    # ACCESS_TOKEN_SECRET = os.getenv("ACCESS_TOKEN_SECRET", None)

    USERS = ['@jarbasAi']

    LANGUAGES = ['en']

    def __init__(self, consumer_key, consumer_secret, access_token,
                 access_token_secret, users, languages=None,
                 name="twitter bridge", *args, **kwargs):
        super(JarbasTwitterBridge, self).__init__(*args, **kwargs)
        self.status = "disconnected"
        self.name = name
        self.client = None
        self.CONSUMER_KEY = consumer_key
        self.CONSUMER_SECRET = consumer_secret
        self.ACCESS_TOKEN = access_token
        self.ACCESS_TOKEN_SECRET = access_token_secret
        self.USERS = users
        self.LANGUAGES = languages or ["en"]

    # websocket handlers
    def clientConnectionFailed(self, connector, reason):
        logger.info(
            "Client connection failed: " + str(reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)

    def clientConnectionLost(self, connector, reason):
        logger.info(
            "Client connection lost: " + str(reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)






# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import time
from threading import Thread

import speech_recognition as sr
from pyee import EventEmitter
from requests import RequestException, HTTPError
from requests.exceptions import ConnectionError

from jarbas_hive_mind.terminals.speech.hotword_factory import HotWordFactory
from jarbas_hive_mind.terminals.speech.mic import MutableMicrophone, \
    ResponsiveRecognizer
from jarbas_hive_mind.terminals.speech.stt import STTFactory

if sys.version_info[0] < 3:
    from Queue import Queue, Empty
else:
    from queue import Queue, Empty

import logging

logger = logging.getLogger("listener")
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel("INFO")

conf = {
    "listener": {
        "sample_rate": 16000,
        "channels": 1,
        "record_wake_words": False,
        "record_utterances": False,
        "phoneme_duration": 120,
        "multiplier": 1.0,
        "energy_ratio": 1.5,
        "wake_word": "hey mycroft",
        "stand_up_word": "wake up"
    },
    "stt": {
        "deepspeech_server": {
            "uri": "http://localhost:8080/stt"
        },
        "kaldi": {
            "uri": "http://localhost:8080/client/dynamic/recognize"
        }
    }
}


class AudioProducer(Thread):
    """
    AudioProducer
    given a mic and a recognizer implementation, continuously listens to the
    mic for potential speech chunks and pushes them onto the queue.
    """

    def __init__(self, state, queue, mic, recognizer, emitter):
        super(AudioProducer, self).__init__()
        self.daemon = True
        self.state = state
        self.queue = queue
        self.mic = mic
        self.recognizer = recognizer
        self.emitter = emitter

    def run(self):
        with self.mic as source:
            self.recognizer.adjust_for_ambient_noise(source)
            while self.state.running:
                try:
                    audio = self.recognizer.listen(source, self.emitter)
                    self.queue.put(audio)
                except IOError as e:
                    # NOTE: Audio stack on raspi is slightly different, throws
                    # IOError every other listen, almost like it can't handle
                    # buffering audio between listen loops.
                    # The internet was not helpful.
                    # http://stackoverflow.com/questions/10733903/pyaudio-input-overflowed
                    self.emitter.emit("recognizer_loop:ioerror", e)

    def stop(self):
        """
            Stop producer thread.
        """
        self.state.running = False
        self.recognizer.stop()


class AudioConsumer(Thread):
    """
    AudioConsumer
    Consumes AudioData chunks off the queue
    """

    # In seconds, the minimum audio size to be sent to remote STT
    MIN_AUDIO_SIZE = 0.5

    def __init__(self, state, queue, emitter, stt,
                 wakeup_recognizer, wakeword_recognizer):
        super(AudioConsumer, self).__init__()
        self.daemon = True
        self.queue = queue
        self.state = state
        self.emitter = emitter
        self.stt = stt
        self.wakeup_recognizer = wakeup_recognizer
        self.wakeword_recognizer = wakeword_recognizer
        self.word = self.wakeword_recognizer.key_phrase
        self.emitter.on("recognizer_loop:hotword", self._set_word)

    def _set_word(self, event):
        self.word = event.get("hotword", self.wakeword_recognizer.key_phrase)

    def run(self):
        while self.state.running:
            self.read()

    def read(self):
        try:
            audio = self.queue.get(timeout=0.5)
        except Empty:
            return

        if audio is None:
            return

        if self.state.sleeping:
            self.wake_up(audio)
        else:
            self.process(audio)

    # TODO: Localization
    def wake_up(self, audio):
        if self.wakeup_recognizer.found_wake_word(audio.frame_data):
            self.state.sleeping = False
            self.emitter.emit('recognizer_loop:awoken')

    @staticmethod
    def _audio_length(audio):
        return float(len(audio.frame_data)) / (
                audio.sample_rate * audio.sample_width)

    # TODO: Localization
    def process(self, audio):

        payload = {
            'utterance': self.word
        }
        self.emitter.emit("recognizer_loop:wakeword", payload)

        if self._audio_length(audio) < self.MIN_AUDIO_SIZE:
            logger.error("[ERROR] Audio too short to be processed")
        else:
            transcription = self.transcribe(audio)
            if transcription:
                # STT succeeded, send the transcribed speech on for processing
                payload = {
                    'utterances': [transcription],
                    'lang': self.stt.lang
                }
                self.emitter.emit("recognizer_loop:utterance", payload)

    def transcribe(self, audio):
        try:
            # Invoke the STT engine on the audio clip
            text = self.stt.execute(audio).lower().strip()
            logger.info("[INFO] STT: " + text)
            return text
        except sr.RequestError as e:
            logger.error(
                "[ERROR] Could not request Speech Recognition {0}".format(e))
        except ConnectionError as e:
            logger.error("[ERROR] Connection Error: {0}".format(e))

            self.emitter.emit("recognizer_loop:no_internet")
        except HTTPError as e:
            logger.error("[ERROR] " + e.__class__.__name__ + ': ' + str(e))
        except RequestException as e:
            logger.error("[ERROR] " + e.__class__.__name__ + ': ' + str(e))
        except Exception as e:
            self.emitter.emit('recognizer_loop:speech.recognition.unknown')
            if isinstance(e, IndexError):
                logger.error('[ERROR] no words were transcribed')
            else:
                logger.exception(e)
            logger.error(
                "[ERROR] Speech Recognition could not understand audio")
            return None
        dialog_name = 'not connected to the internet'
        self.emitter.emit('speak', {'utterance': dialog_name})

    def __speak(self, utterance):
        payload = {
            'utterance': utterance
        }
        self.emitter.emit("speak", payload)


class RecognizerLoopState(object):
    def __init__(self):
        self.running = False
        self.sleeping = False


class RecognizerLoop(EventEmitter):
    """
        EventEmitter loop running speech recognition. Local wake word
        recognizer and remote general speech recognition.
    """

    def __init__(self, config=None):
        super(RecognizerLoop, self).__init__()
        self.mute_calls = 0
        self._load_config(config)

    def _load_config(self, config=None):
        """
            Load configuration parameters from configuration
        """
        config = config or conf
        self.config_core = config
        self._config_hash = hash(str(config))
        self.lang = config.get('lang')
        self.config = config.get('listener')
        rate = self.config.get('sample_rate')
        device_index = self.config.get('device_index')

        self.microphone = MutableMicrophone(device_index, rate,
                                            mute=self.mute_calls > 0)
        # FIXME - channels are not been used
        self.microphone.CHANNELS = self.config.get('channels')
        self.wakeword_recognizer = self.create_wake_word_recognizer()
        # TODO - localization
        self.wakeup_recognizer = self.create_wakeup_recognizer()
        self.hot_word_engines = {}
        self.create_hot_word_engines()
        self.responsive_recognizer = ResponsiveRecognizer(
            self.wakeword_recognizer, self.hot_word_engines)
        self.state = RecognizerLoopState()

    def create_hot_word_engines(self):
        logger.info("[INFO] creating secondary hotword engines")
        hot_words = self.config_core.get("hotwords", {})
        for word in hot_words:
            data = hot_words[word]
            if word == self.wakeup_recognizer.key_phrase \
                    or word == self.wakeword_recognizer.key_phrase \
                    or not data.get("active", False):
                continue
            engine_type = data["module"]
            ding = data.get("sound")
            utterance = data.get("utterance")
            listen = data.get("listen", False)
            engine = HotWordFactory.create_hotword(word, lang=self.lang)
            self.hot_word_engines[word] = [engine, ding, utterance,
                                           listen, engine_type]

    def create_wake_word_recognizer(self):
        # Create a local recognizer to hear the wakeup word, e.g. 'Hey Mycroft'
        logger.info("[INFO] creating main wake word engine")
        word = self.config.get("wake_word", "hey mycroft")
        # TODO remove this, only for server settings compatibility
        phonemes = self.config.get("phonemes")
        thresh = self.config.get("threshold")
        config = self.config_core.get("hotwords", {word: {}})

        if word not in config:
            config[word] = {'module': 'pocketsphinx'}
        if phonemes:
            config[word]["phonemes"] = phonemes
        if thresh:
            config[word]["threshold"] = thresh
        if phonemes is None or thresh is None:
            config = None
        return HotWordFactory.create_hotword(word, config, self.lang)

    def create_wakeup_recognizer(self):
        logger.info("[INFO] creating stand up word engine")
        word = self.config.get("stand_up_word", "wake up")
        return HotWordFactory.create_hotword(word, lang=self.lang)

    def start_async(self):
        """
            Start consumer and producer threads
        """
        self.state.running = True
        queue = Queue()
        self.producer = AudioProducer(self.state, queue, self.microphone,
                                      self.responsive_recognizer, self)
        self.producer.start()
        self.consumer = AudioConsumer(self.state, queue, self,
                                      STTFactory.create(),
                                      self.wakeup_recognizer,
                                      self.wakeword_recognizer)
        self.consumer.start()

    def stop(self):
        self.state.running = False
        self.producer.stop()
        # wait for threads to shutdown
        self.producer.join()
        self.consumer.join()

    def mute(self):
        """
            Mute microphone and increase number of requests to mute
        """
        self.mute_calls += 1
        if self.microphone:
            self.microphone.mute()

    def unmute(self):
        """
            Unmute mic if as many unmute calls as mute calls have been
            received.
        """
        if self.mute_calls > 0:
            self.mute_calls -= 1

        if self.mute_calls <= 0 and self.microphone:
            self.microphone.unmute()
            self.mute_calls = 0

    def force_unmute(self):
        """
            Completely unmute mic dispite the number of calls to mute
        """
        self.mute_calls = 0
        self.unmute()

    def is_muted(self):
        if self.microphone:
            return self.microphone.is_muted()
        else:
            return True  # consider 'no mic' muted

    def sleep(self):
        self.state.sleeping = True

    def awaken(self):
        self.state.sleeping = False

    def run(self):
        self.start_async()
        while self.state.running:
            try:
                time.sleep(1)

            except KeyboardInterrupt as e:
                logger.exception(e)
                self.stop()
                raise  # Re-raise KeyboardInterrupt

    def reload(self):
        """
            Reload configuration and restart consumer and producer
        """
        self.stop()
        # load config
        self._load_config()
        # restart
        self.start_async()



# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import os.path
import tempfile
import time


def get_ipc_directory(domain=None):
    """Get the directory used for Inter Process Communication

    Files in this folder can be accessed by different processes on the
    machine.  Useful for communication.  This is often a small RAM disk.

    Args:
        domain (str): The IPC domain.  Basically a subdirectory to prevent
            overlapping signal filenames.

    Returns:
        str: a path to the IPC directory
    """
    dir = os.path.join(tempfile.gettempdir(), "mycroft", "ipc")
    return ensure_directory_exists(dir, domain)


def ensure_directory_exists(directory, domain=None):
    """ Create a directory and give access rights to all

    Args:
        domain (str): The IPC domain.  Basically a subdirectory to prevent
            overlapping signal filenames.

    Returns:
        str: a path to the directory
    """
    if domain:
        directory = os.path.join(directory, domain)
    directory = os.path.normpath(directory)

    if not os.path.isdir(directory):
        try:
            save = os.umask(0)
            os.makedirs(directory, 0o777)  # give everyone rights to r/w here
        except OSError:
            print("Failed to create: " + directory)
            pass
        finally:
            os.umask(save)

    return directory


def create_file(filename):
    """ Create the file filename and create any directories needed

        Args:
            filename: Path to the file to be created
    """
    try:
        os.makedirs(os.path.dirname(filename))
    except OSError:
        pass
    with open(filename, 'w') as f:
        f.write('')


def create_signal(signal_name):
    """Create a named signal

    Args:
        signal_name (str): The signal's name.  Must only contain characters
            valid in filenames.
    """
    try:
        path = os.path.join(get_ipc_directory(), "signal", signal_name)
        create_file(path)
        return os.path.isfile(path)
    except IOError:
        return False


def check_for_signal(signal_name, sec_lifetime=0):
    """See if a named signal exists

    Args:
        signal_name (str): The signal's name.  Must only contain characters
            valid in filenames.
        sec_lifetime (int, optional): How many seconds the signal should
            remain valid.  If 0 or not specified, it is a single-use signal.
            If -1, it never expires.

    Returns:
        bool: True if the signal is defined, False otherwise
    """
    path = os.path.join(get_ipc_directory(), "signal", signal_name)
    if os.path.isfile(path):
        if sec_lifetime == 0:
            # consume this single-use signal
            os.remove(path)
        elif sec_lifetime == -1:
            return True
        elif int(os.path.getctime(path) + sec_lifetime) < int(time.time()):
            # remove once expired
            os.remove(path)
            return False
        return True

    # No such signal exists
    return False



import json
import logging
import sys
from threading import Thread
from time import sleep

from autobahn.twisted.websocket import WebSocketClientFactory, \
    WebSocketClientProtocol
from fbchat import log, Client, ThreadType
from fbchat.utils import Message
from twisted.internet.protocol import ReconnectingClientFactory

platform = "JarbasFaceBookBridgev0.1"
logger = logging.getLogger(platform)
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel("INFO")


class FaceBot(Client):
    protocol = None

    def bind(self, protocol):
        self.protocol = protocol

    def onMessage(self, author_id, message_object, thread_id, thread_type,
                  **kwargs):
        self.markAsDelivered(author_id, thread_id)
        self.markAsRead(author_id)

        log.info("{} from {} in {}".format(message_object, thread_id,
                                           thread_type.name))

        # If you're not the author, process
        if author_id != self.uid:
            utterance = message_object.text
            if thread_type == ThreadType.GROUP:
                name = self.fetchUserInfo(self.uid)[self.uid].name
                if "@" + name.lower() not in utterance.lower():
                    return
                utterance = utterance.replace("@" + name, "")
            msg = {"data": {"utterances": [utterance],
                            "lang": "en-us"},
                   "type": "recognizer_loop:utterance",
                   "context": {"source": self.protocol.peer,
                               "destination": "hive_mind",
                               "platform": platform,
                               "user": author_id,
                               "fb_chat_id": thread_id,
                               "thread_type": thread_type.name}}
            msg = json.dumps(msg)
            msg = bytes(msg, encoding="utf-8")
            self.protocol.sendMessage(msg, False)


class JarbasFacebookBridgeProtocol(WebSocketClientProtocol):
    facebook = None
    chat_thread = None
    MAIL = ""
    PASSWD = ""

    def start_fb_chat(self):
        self.facebook = FaceBot(self.MAIL, self.PASSWD)
        self.facebook.listen()

    def onConnect(self, response):
        logger.info("[INFO] Server connected: {0}".format(response.peer))
        self.factory.client = self
        self.factory.status = "connected"
        self.MAIL = self.factory.mail
        self.PASSWD = self.factory.passwd

    def onOpen(self):
        logger.info("[INFO] WebSocket connection open. ")

        self.chat_thread = Thread(target=self.start_fb_chat)
        self.chat_thread.setDaemon(True)
        self.chat_thread.start()
        while self.facebook is None:
            sleep(1)
        self.facebook.bind(self)

    def onMessage(self, payload, isBinary):
        if not isBinary:
            payload = payload.decode("utf-8")
            msg = json.loads(payload)
            utterance = ""
            if msg.get("type", "") == "speak":
                utterance = msg["data"]["utterance"]
            elif msg.get("type", "") == "server.complete_intent_failure":
                utterance = "does not compute"

            if utterance:
                user_id = msg["context"]["fb_chat_id"]
                thread_type = msg["context"]["thread_type"]
                if thread_type == "GROUP":
                    thread_type = ThreadType.GROUP
                else:
                    thread_type = ThreadType.USER
                self.facebook.send(Message(text=utterance),
                                   thread_id=user_id,
                                   thread_type=thread_type)
        else:
            pass

    def onClose(self, wasClean, code, reason):
        logger.info("[INFO] WebSocket connection closed: {0}".format(reason))
        self.factory.client = None
        self.factory.status = "disconnected"


class JarbasFacebookBridge(WebSocketClientFactory,
                           ReconnectingClientFactory):
    protocol = JarbasFacebookBridgeProtocol

    def __init__(self, mail, password, name="facebook bridge", *args,
                 **kwargs):
        super(JarbasFacebookBridge, self).__init__(*args, **kwargs)
        self.status = "disconnected"
        self.name = name
        self.client = None
        self.mail = mail
        self.passwd = password

    # websocket handlers
    def clientConnectionFailed(self, connector, reason):
        logger.info(
            "Client connection failed: " + str(reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)

    def clientConnectionLost(self, connector, reason):
        logger.info(
            "Client connection lost: " + str(reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)





import json
from threading import Thread

from autobahn.twisted.websocket import WebSocketClientFactory, \
    WebSocketClientProtocol
from twisted.internet.protocol import ReconnectingClientFactory

from jarbas_hive_mind.utils.log import LOG as logger
from jarbas_hive_mind.utils.messagebus.message import Message
from jarbas_hive_mind.utils.messagebus.ws import WebsocketClient

platform = "Jarbas Drone"


class JarbasDroneProtocol(WebSocketClientProtocol):

    def onConnect(self, response):
        logger.info("Server connected: {0}".format(response.peer))
        self.factory.bus.emit(Message("hive.mind.connected",
                                      {"server_id": response.headers[
                                              "server"]}))
        self.factory.client = self
        self.factory.status = "connected"

    def onOpen(self):
        logger.info("WebSocket connection open. ")
        self.factory.bus.emit(Message("hive.mind.websocket.open"))

    def onMessage(self, payload, isBinary):
        logger.info("status: " + self.factory.status)
        if not isBinary:
            payload = payload.decode("utf-8")
            data = {"payload": payload, "isBinary": isBinary}
        else:
            data = {"payload": None, "isBinary": isBinary}
        self.factory.bus.emit(Message("hive.mind.message.received",
                                      data))

    def onClose(self, wasClean, code, reason):
        logger.info("WebSocket connection closed: {0}".format(reason))
        self.factory.bus.emit(Message("hive.mind.connection.closed",
                                      {"wasClean": wasClean,
                                           "reason": reason,
                                           "code": code}))
        self.factory.client = None
        self.factory.status = "disconnected"

    def serialize_message(self, message):
        # convert a Message object into raw data that can be sent over
        # websocket
        if hasattr(message, 'serialize'):
            return message.serialize()
        else:
            return json.dumps(message.__dict__)


class JarbasDrone(WebSocketClientFactory, ReconnectingClientFactory):
    protocol = JarbasDroneProtocol

    def __init__(self, bus=None, *args, **kwargs):
        super(JarbasDrone, self).__init__(*args, **kwargs)
        self.client = None
        self.status = "disconnected"
        # mycroft_ws
        self.bus = bus
        self.mycroft_bus_thread = None
        self.create_mycroft_bus()

    # initialize methods
    def connect_to_mycroft(self):
        self.bus.run_forever()

    def create_mycroft_bus(self):
        # connect to mycroft internal websocket
        self.bus = self.bus or WebsocketClient()
        self.register_mycroft_messages()
        self.mycroft_bus_thread = Thread(target=self.connect_to_mycroft)
        self.mycroft_bus_thread.setDaemon(True)
        self.mycroft_bus_thread.start()

    def register_mycroft_messages(self):
        self.bus.on("hive.mind.message.received",
                    self.handle_receive_server_message)
        self.bus.on("hive.mind.message.send",
                    self.handle_send_server_message)

    def shutdown(self):
        self.bus.remove("hive.mind.message.received",
                        self.handle_receive_server_message)
        self.bus.remove("hive.mind.message.send",
                        self.handle_send_server_message)

    # websocket handlers
    def clientConnectionFailed(self, connector, reason):
        logger.info(
            "Client connection failed: " + str(reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)

    def clientConnectionLost(self, connector, reason):
        logger.info(
            "Client connection lost: " + str(reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)

    # mycroft handlers
    def handle_receive_server_message(self, message):
        server_msg = message.data.get("payload")
        is_file = message.data.get("isBinary")
        if is_file:
            # TODO received file
            pass
        else:
            # forward server message to internal bus
            message = Message.deserialize(server_msg)
            self.bus.emit(message)

    def handle_send_server_message(self, message):
        server_msg = message.data.get("payload")
        is_file = message.data.get("isBinary")
        if is_file:
            # TODO send file
            pass
        else:
            # send message to server
            server_msg = Message.deserialize(server_msg)
            server_msg.context["platform"] = platform
            self.sendMessage(server_msg.type, server_msg.data,
                             server_msg.context)

    def sendRaw(self, data):
        if self.client is None:
            logger.error("Client is none")
            return
        self.client.sendMessage(data, isBinary=True)

    def sendMessage(self, type, data, context=None):
        if self.client is None:
            logger.error("Client is none")
            return
        if context is None:
            context = {}
        msg = self.client.serialize_message(Message(type, data, context))
        self.client.sendMessage(msg, isBinary=False)
        self.bus.emit(Message("hive.mind.message.sent",
                              {"type": type,
                               "data": data,
                               "context": context,
                               "raw": msg}))





from os import makedirs
from os.path import isdir, join, expanduser

DATA_PATH = expanduser("~/jarbas_hivemind")

path_hierarcy = [
    "~/.mycroft/hivemind",
    "~/.jarbas/hivemind"
    "~/jarbas_hivemind",
    "~/mycroft/hivemind",
    "~/jarbas/hivemind",
]

for p in path_hierarcy:
    if isdir(p):
        DATA_PATH = expanduser(p)
        break

if not isdir(DATA_PATH):
    makedirs(DATA_PATH)

CERTS_PATH = join(DATA_PATH, "certs")
if not isdir(CERTS_PATH):
    makedirs(CERTS_PATH)

DEFAULT_SSL_CRT = join(CERTS_PATH, "hivemind.crt")
DEFAULT_SSL_KEY = join(CERTS_PATH, "hivemind.key")

DB_PATH = join(DATA_PATH, "database")
if not isdir(DB_PATH):
    makedirs(DB_PATH)

CLIENTS_DB = "sqlite:///" + join(DB_PATH, "clients.db")

DEFAULT_PORT = 5678
USE_SSL = False

LOG_BLACKLIST = []

MYCROFT_WEBSOCKET_CONFIG = {
    "host": "0.0.0.0",
    "port": 8181,
    "route": "/core",
    "ssl": False
}

STT_CONFIG = {"stt": {
    "module": "google",
    "deepspeech_server": {
        "uri": "http://localhost:8080/stt"
    },
    "kaldi": {
        "uri": "http://localhost:8080/client/dynamic/recognize"
    }
}}

LISTENER_CONFIG = {
    "listener": {
        "sample_rate": 16000,
        "channels": 1,
        "record_wake_words": False,
        "record_utterances": False,
        "phoneme_duration": 120,
        "multiplier": 1.0,
        "energy_ratio": 1.5,
        "wake_word": "hey mycroft",
        "stand_up_word": "wake up"
    },
    "hotwords": {
        "hey mycroft": {
            "module": "pocketsphinx",
            "phonemes": "HH EY . M AY K R AO F T",
            "threshold": 1e-90,
            "lang": "en-us"
        },
        "thank you": {
            "module": "pocketsphinx",
            "phonemes": "TH AE NG K . Y UW .",
            "threshold": 1e-1,
            "listen": False,
            "utterance": "thank you",
            "active": False,
            "sound": "",
            "lang": "en-us"
        },
        "wake up": {
            "module": "pocketsphinx",
            "phonemes": "W EY K . AH P",
            "threshold": 1e-20,
            "lang": "en-us"
        }
    }
}



import base64
import json
import logging
import random
from threading import Thread

from autobahn.twisted.websocket import WebSocketClientFactory, \
    WebSocketClientProtocol
from remi import start, App, gui
from twisted.internet import reactor, ssl
from twisted.internet.protocol import ReconnectingClientFactory

platform = "JarbasREMITerminalv0.1"

logger = logging.getLogger(platform)
logger.setLevel("INFO")


class JarbasRemiTerminalProtocol(WebSocketClientProtocol):
    remi = None

    def onConnect(self, response):
        logger.info("[INFO] Server connected: {0}".format(response.peer))
        self.factory.client = self
        self.factory.status = "connected"

    def onOpen(self):
        logger.info("[INFO] WebSocket connection open. ")
        RemiTerminal.protocol = self

    def onMessage(self, payload, isBinary):
        if not isBinary:
            payload = payload.decode("utf-8")
            msg = json.loads(payload)
            if msg.get("type", "") == "speak":
                utterance = msg["data"]["utterance"]
                logger.info("[OUTPUT] " + utterance)
                RemiTerminal.history_widget.append(
                    "Jarbas: " + utterance.lower())
            elif msg.get("type", "") == "hive.complete_intent_failure":
                logger.error("[ERROR] complete intent failure")
        else:
            pass

    def onClose(self, wasClean, code, reason):
        logger.info("[INFO] WebSocket connection closed: {0}".format(reason))
        self.factory.client = None
        self.factory.status = "disconnected"
        if "Internalservererror:InvalidAPIkey" in reason:
            logger.error("[ERROR] invalid user:key provided")
            # TODO show something in UI ?
            raise ConnectionAbortedError("invalid user:key provided")


class JarbasRemiTerminal(WebSocketClientFactory,
                         ReconnectingClientFactory):
    protocol = JarbasRemiTerminalProtocol

    def __init__(self, *args, **kwargs):
        super(JarbasRemiTerminal, self).__init__(*args, **kwargs)
        self.status = "disconnected"
        self.client = None

    # websocket handlers
    def clientConnectionFailed(self, connector, reason):
        logger.info("[INFO] Client connection failed: " + str(reason) +
                    " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)

    def clientConnectionLost(self, connector, reason):
        logger.info("[INFO] Client connection lost: " + str(reason) +
                    " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)


class RemiTerminal(App):
    protocol = None
    history_widget = None
    suggestions = ["hello world",
                   "do you like pizza",
                   "tell me about nicola tesla",
                   "tell me a joke"]
    host = "127.0.0.1"
    port = 5678
    name = "Jarbas Remi Terminal"
    key = "remi_key"
    authorization = bytes(name + ":" + key, encoding="utf-8")
    usernamePasswordDecoded = authorization
    key = base64.b64encode(usernamePasswordDecoded)

    def __init__(self, *args):
        super(RemiTerminal, self).__init__(*args)

    def main(self):
        authorization = bytes(self.name + ":" + self.key, encoding="utf-8")
        usernamePasswordDecoded = authorization
        key = base64.b64encode(usernamePasswordDecoded)

        headers = {'authorization': key}
        adress = u"wss://" + self.host + u":" + str(self.port)
        terminal = JarbasRemiTerminal(adress, headers=headers,
                                      useragent=platform)
        contextFactory = ssl.ClientContextFactory()
        reactor.connectSSL(self.host, self.port, terminal, contextFactory)

        self.reactor_loop = Thread(target=reactor.run)
        self.reactor_loop.setDaemon(True)
        self.reactor_loop.start()

        # returning the root widget
        return self.get_chat_widget()

    def get_chat_widget(self):
        verticalContainer = gui.Widget(width=400, margin='0px auto',
                                       style={'display': 'block',
                                              'overflow': 'hidden'})
        chatButtonContainer = gui.Widget(width=400,
                                         layout_orientation=gui.Widget.LAYOUT_HORIZONTAL,
                                         margin='0px',
                                         style={'display': 'block',
                                                'overflow': 'auto'})

        RemiTerminal.history_widget = gui.ListView.new_from_list((), width=500,
                                                                 height=300,
                                                                 margin='10px')

        self.txt_input = gui.TextInput(width=400, height=30, margin='10px')
        self.txt_input.set_on_change_listener(self.on_chat_enter)

        send_button = gui.Button('Send', width=150, height=30, margin='10px')
        send_button.set_on_click_listener(self.on_chat_click)

        sug_button = gui.Button('Suggestion', width=150, height=30,
                                margin='10px')
        sug_button.set_on_click_listener(self.on_suggestion_click)

        chatButtonContainer.append(send_button)
        chatButtonContainer.append(sug_button)

        verticalContainer.append(self.txt_input)
        verticalContainer.append(chatButtonContainer)
        verticalContainer.append(RemiTerminal.history_widget)
        return verticalContainer

    def on_suggestion_click(self, widget):
        sug = random.choice(self.suggestions)
        self.txt_input.set_text(sug)
        self.utterance = sug

    def on_chat_type(self, widget, newValue):
        self.utterance = str(newValue)

    def on_chat_click(self, widget):
        self.utterance = self.utterance.strip()
        if self.utterance:
            msg = {"data": {"utterances": [self.utterance], "lang": "en-us"},
                   "type": "recognizer_loop:utterance",
                   "context": {"source": RemiTerminal.protocol.peer,
                               "destination": "hive_mind",
                               "platform": platform}}
            msg = json.dumps(msg)
            msg = bytes(msg, encoding="utf-8")
            RemiTerminal.protocol.sendMessage(msg, False)

            RemiTerminal.history_widget.append(
                "you: " + self.utterance.strip())
            self.txt_input.set_text('')
            self.utterance = ""

    def on_chat_enter(self, widget, userData):
        self.utterance = userData.strip()
        if self.utterance:
            msg = {"data": {"utterances": [self.utterance], "lang": "en-us"},
                   "type": "recognizer_loop:utterance",
                   "context": {"source": RemiTerminal.protocol.peer,
                               "destination": "hive_mind",
                               "platform": platform}}
            msg = json.dumps(msg)
            msg = bytes(msg, encoding="utf-8")
            RemiTerminal.protocol.sendMessage(msg, False)

            RemiTerminal.history_widget.append(
                "you: " + self.utterance.strip())
            self.txt_input.set_text('')
            self.utterance = ""






import base64
from twisted.internet import reactor, ssl
from jarbas_hive_mind.bridges.twitch import platform, logger, JarbasTwitchBridge


def connect_to_twitch(oauth, channel, username="Jarbas_BOT", host="127.0.0.1",
                      port=5678, name="Jarbas Twitch Bridge",
                      api="twitch_key", useragent=platform):
    authorization = bytes(name + ":" + api, encoding="utf-8")
    usernamePasswordDecoded = authorization
    api = base64.b64encode(usernamePasswordDecoded)
    headers = {'authorization': api}
    address = u"wss://" + host + u":" + str(port)
    logger.info("[INFO] connected to hivemind: " + address)
    bridge = JarbasTwitchBridge(oauth=oauth, channel=channel,
                                username=username, headers=headers,
                                useragent=useragent)
    contextFactory = ssl.ClientContextFactory()
    reactor.connectSSL(host, port, bridge, contextFactory)
    reactor.run()


if __name__ == '__main__':
    BOTNAME = "Jarbas_BOT"
    PASS = "oauth:XXX"
    CHANNELNAME = "jarbasai"
    connect_to_twitch(PASS, CHANNELNAME, BOTNAME)



import json
import logging
import sys
import time

import requests
from requests.exceptions import ConnectionError
# filter warnings, this should be removed once we stop using self signed
# certs for debug
from requests.packages.urllib3.exceptions import InsecureRequestWarning

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

platform = "JarbasHttpsCliTerminalv0.1"
logger = logging.getLogger(platform)
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel("INFO")


class JarbasMicroServicesAPI(object):
    def __init__(self, key, url="https://0.0.0.0:6712/", lang="en-us"):
        self.key = key
        self.headers = {"Authorization": key}
        self.lang = lang
        self.url = url
        self.timeout = 10
        self.wait_time = 0.5

    def hello_world(self):
        try:
            response = requests.get(
                self.url,
                headers=self.headers, verify=False
            )
            return response.text
        except ConnectionError as e:
            raise ConnectionError("Could not connect: " + str(e))

    def new_user(self, key, mail, name):
        ''' add a new user, requires admin api '''
        try:
            response = requests.put(
                self.url + "new_user/" + key + "/" + mail + "/" + name,
                headers=self.headers, verify=False
            )
            try:
                return response.json()
            except:
                logger.error(response.text)
                raise ValueError("Invalid admin api key")
        except ConnectionError as e:
            raise ConnectionError("Could not connect: " + str(e))

    def revoke_key(self, key):
        ''' add a new user, requires admin api '''
        try:
            response = requests.put(
                self.url + "revoke_key/" + key,
                headers=self.headers, verify=False
            )
            try:
                return response.json()
            except:
                logger.error(response.text)
                raise ValueError("Invalid admin api key")
        except ConnectionError as e:
            raise ConnectionError("Could not connect: " + str(e))

    def generate_key(self):
        ''' get an api key string, requires admin api '''
        try:
            response = requests.get(
                self.url + "generate_key",
                headers=self.headers, verify=False
            )
            try:
                return response.json()["key"]
            except:
                logger.info(response.text)
                raise ValueError("Invalid admin api key")
        except ConnectionError as e:
            raise ConnectionError("Could not connect: " + str(e))

    def ask_mycroft(self, utterance, lang=None):
        ''' ask something to mycroft '''
        lang = lang or self.lang
        try:
            response = requests.put(
                self.url + "ask/" + lang + "/" + utterance,
                headers=self.headers, verify=False
            )
            try:
                ans = response.json()
                if ans["status"] == "processing":
                    # start waiting
                    start = time.time()
                    while ans["status"] == "processing":
                        time.sleep(self.wait_time)
                        if time.time() - start > self.timeout:
                            try:
                                response = requests.put(
                                    self.url + "cancel",
                                    headers=self.headers, verify=False
                                )
                            except Exception as e:
                                logger.error(e)

                            return {"type": "speak",
                                    "data": {"utterance": "server timed out"},
                                    "context": {"source": "hivemind_flask",
                                                "destination": self.key}}
                        try:
                            response = requests.get(
                                self.url + "get_answer",
                                headers=self.headers, verify=False
                            )
                            ans = response.json()
                        except Exception as e:
                            raise ValueError("Unexpected Error: " + str(e))
                    return json.loads(ans["answer"])
                else:
                    raise ValueError("Received unexpected status from "
                                     "server: " + str(ans))
            except:
                logger.error(response.text)
                raise ValueError("Invalid api key: " + str(self.key))
        except ConnectionError as e:
            raise ConnectionError("Could not connect: " + str(e))




# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import os
import sys
import tempfile
import time
from os import mkdir, getcwd, chdir
from os.path import dirname, exists, join, abspath, expanduser, isdir, isfile
from subprocess import Popen, PIPE, call
from tempfile import NamedTemporaryFile
from threading import Thread
from time import time as get_time

logger = logging.getLogger("hotwords")
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel("INFO")

conf = {
    "listener": {
        "sample_rate": 16000,
        "channels": 1,
        "record_wake_words": False,
        "record_utterances": False,
        "phoneme_duration": 120,
        "multiplier": 1.0,
        "energy_ratio": 1.5,
        "wake_word": "hey mycroft",
        "stand_up_word": "wake up"
      },
    "hotwords": {
        "hey mycroft": {
            "module": "pocketsphinx",
            "phonemes": "HH EY . M AY K R AO F T",
            "threshold": 1e-90,
            "lang": "en-us"
        },
        "thank you": {
            "module": "pocketsphinx",
            "phonemes": "TH AE NG K . Y UW .",
            "threshold": 1e-1,
            "listen": False,
            "utterance": "thank you",
            "active": False,
            "sound": "",
            "lang": "en-us"
        },
        "wake up": {
            "module": "pocketsphinx",
            "phonemes": "W EY K . AH P",
            "threshold": 1e-20,
            "lang": "en-us"
        }
    },
    "precise": {
        "dist_url": "https://raw.githubusercontent.com/MycroftAI/precise-data/dist/",
        "models_url": "https://raw.githubusercontent.com/MycroftAI/precise-data/models/"
      }
}


RECOGNIZER_DIR = join(abspath(dirname(__file__)), "recognizer")


class HotWordEngine(object):
    def __init__(self, key_phrase="hey mycroft", config=None, lang="en-us"):
        self.key_phrase = str(key_phrase).lower()
        # rough estimate 1 phoneme per 2 chars
        self.num_phonemes = len(key_phrase) / 2 + 1
        if config is None:
            config = conf.get("hotwords", {})
            config = config.get(self.key_phrase, {})
        self.config = config
        self.listener_config = conf.get("listener", {})
        self.module = self.config.get("module")
        self.lang = str(self.config.get("lang", lang)).lower()

    def found_wake_word(self, frame_data):
        return False

    def update(self, chunk):
        pass


class PocketsphinxHotWord(HotWordEngine):
    def __init__(self, key_phrase="hey mycroft", config=None, lang="en-us"):
        super(PocketsphinxHotWord, self).__init__(key_phrase, config, lang)
        # Hotword module imports
        try:
            # pip install
            from pocketsphinx import Decoder
        except ImportError:
            # source install
            from pocketsphinx.pocketsphinx import Decoder
        # Hotword module params
        self.phonemes = self.config.get("phonemes", "HH EY . M AY K R AO F T")
        self.num_phonemes = len(self.phonemes.split())
        self.threshold = self.config.get("threshold", 1e-90)
        self.sample_rate = self.listener_config.get("sample_rate", 1600)
        dict_name = self.create_dict(self.key_phrase, self.phonemes)
        config = self.create_config(dict_name, Decoder.default_config())
        self.decoder = Decoder(config)

    def create_dict(self, key_phrase, phonemes):
        (fd, file_name) = tempfile.mkstemp()
        words = key_phrase.split()
        phoneme_groups = phonemes.split('.')
        with os.fdopen(fd, 'w') as f:
            for word, phoneme in zip(words, phoneme_groups):
                f.write(word + ' ' + phoneme + '\n')
        return file_name

    def create_config(self, dict_name, config):
        model_file = join(RECOGNIZER_DIR, 'model', self.lang, 'hmm')
        if not exists(model_file):
            logger.error(
                '[ERROR] PocketSphinx model not found at ' + str(model_file))
        config.set_string('-hmm', model_file)
        config.set_string('-dict', dict_name)
        config.set_string('-keyphrase', self.key_phrase)
        config.set_float('-kws_threshold', float(self.threshold))
        config.set_float('-samprate', self.sample_rate)
        config.set_int('-nfft', 2048)
        config.set_string('-logfn', '/dev/null')
        return config

    def transcribe(self, byte_data, metrics=None):
        start = time.time()
        self.decoder.start_utt()
        self.decoder.process_raw(byte_data, False, False)
        self.decoder.end_utt()
        if metrics:
            metrics.timer("mycroft.stt.local.time_s", time.time() - start)
        return self.decoder.hyp()

    def found_wake_word(self, frame_data):
        hyp = self.transcribe(frame_data)
        return hyp and self.key_phrase in hyp.hypstr.lower()


class PreciseHotword(HotWordEngine):
    def __init__(self, key_phrase="hey mycroft", config=None, lang="en-us"):
        super(PreciseHotword, self).__init__(key_phrase, config, lang)
        self.update_freq = 24  # in hours
        precise_config = config.get('precise')
        self.dist_url = precise_config['dist_url']
        self.models_url = precise_config['models_url']
        self.exe_name = 'precise-stream'

        model_name, model_path = self.get_model_info()

        exe_file = self.find_download_exe()
        logger.info('[INFO] Found precise executable: ' + exe_file)
        self.update_model(model_name, model_path)

        args = [exe_file, model_path, '1024']
        self.proc = Popen(args, stdin=PIPE, stdout=PIPE)
        self.has_found = False
        self.cooldown = 20
        t = Thread(target=self.check_stdout)
        t.daemon = True
        t.start()

    def get_model_info(self):
        ww = conf['listener']['wake_word']
        model_name = ww.replace(' ', '-') + '.pb'
        model_folder = expanduser('~/.mycroft/precise')
        if not isdir(model_folder):
            mkdir(model_folder)
        model_path = join(model_folder, model_name)
        return model_name, model_path

    def find_download_exe(self):
        try:
            if call('command -v ' + self.exe_name,
                    shell=True, stdout=PIPE) == 0:
                return self.exe_name
        except OSError:
            pass

        precise_folder = expanduser('~/.mycroft/precise')
        if isfile(join(precise_folder, 'precise-stream')):
            os.remove(join(precise_folder, 'precise-stream'))

        exe_file = join(precise_folder, 'precise-stream', 'precise-stream')
        if isfile(exe_file):
            return exe_file

        import platform
        import stat

        def snd_msg(cmd):
            """Send message to faceplate"""
            Popen('echo "' + cmd + '" > /dev/ttyAMA0', shell=True)

        arch = platform.machine()

        url = self.dist_url + arch + '/precise-stream.tar.gz'
        tar_file = NamedTemporaryFile().name + '.tar.gz'

        snd_msg('mouth.text=Updating Listener...')
        cur_dir = getcwd()
        chdir(precise_folder)
        try:
            self.download(url, tar_file)
            call(['tar', '-xzvf', tar_file])
        finally:
            chdir(cur_dir)
            snd_msg('mouth.reset')

        if not isfile(exe_file):
            raise RuntimeError('Could not extract file: ' + exe_file)
        os.chmod(exe_file, os.stat(exe_file).st_mode | stat.S_IEXEC)
        return exe_file

    @staticmethod
    def download(url, filename):
        import shutil

        # python 2/3 compatibility
        if sys.version_info[0] >= 3:
            from urllib.request import urlopen
        else:
            from urllib2 import urlopen
        logger.info('[INFO] Downloading: ' + url)
        req = urlopen(url)
        with open(filename, 'wb') as fp:
            shutil.copyfileobj(req, fp)
        logger.info('[INFO] Download complete.')

    def update_model(self, name, file_name):
        if isfile(file_name):
            stat = os.stat(file_name)
            if get_time() - stat.st_mtime < self.update_freq * 60 * 60:
                return
        name = name.replace(' ', '%20')
        url = self.models_url + name
        self.download(url, file_name)
        self.download(url + '.params', file_name + '.params')

    def check_stdout(self):
        while True:
            line = self.proc.stdout.readline()
            if self.cooldown > 0:
                self.cooldown -= 1
                self.has_found = False
                continue
            self.has_found = float(line) > 0.5

    def update(self, chunk):
        self.proc.stdin.write(chunk)
        self.proc.stdin.flush()

    def found_wake_word(self, frame_data):
        if self.has_found and self.cooldown == 0:
            self.cooldown = 20
            return True
        return False


class SnowboyHotWord(HotWordEngine):
    def __init__(self, key_phrase="hey mycroft", config=None, lang="en-us"):
        super(SnowboyHotWord, self).__init__(key_phrase, config, lang)
        # Hotword module imports
        from snowboydecoder import HotwordDetector
        # Hotword params
        models = self.config.get("models", {})
        paths = []
        for key in models:
            paths.append(models[key])
        sensitivity = self.config.get("sensitivity", 0.5)
        self.snowboy = HotwordDetector(paths,
                                       sensitivity=[sensitivity] * len(paths))
        self.lang = str(lang).lower()
        self.key_phrase = str(key_phrase).lower()

    def found_wake_word(self, frame_data):
        wake_word = self.snowboy.detector.RunDetection(frame_data)
        return wake_word == 1


class HotWordFactory(object):
    CLASSES = {
        "pocketsphinx": PocketsphinxHotWord,
        "precise": PreciseHotword,
        "snowboy": SnowboyHotWord
    }

    @staticmethod
    def create_hotword(hotword="hey mycroft", config=None, lang="en-us"):
        logger.info("[INFO] creating " + hotword)
        if not config:
            config = conf.get("hotwords", {})
        module = config.get(hotword).get("module", "pocketsphinx")
        config = config.get(hotword, {"module": module})
        clazz = HotWordFactory.CLASSES.get(module)
        try:
            return clazz(hotword, config, lang=lang)
        except Exception:
            logger.warning(
                '[WARNING] Could not create hotword. Falling back to default.')
            return HotWordFactory.CLASSES['pocketsphinx']()



from jarbas_hive_mind.minds.simple import JarbasMind, JarbasMindProtocol
from twisted.internet import reactor, ssl
from jarbas_hive_mind.settings import CERTS_PATH, DEFAULT_SSL_CRT, \
    DEFAULT_SSL_KEY, DEFAULT_PORT, USE_SSL
from jarbas_hive_mind.utils import create_self_signed_cert
from jarbas_hive_mind.utils.log import LOG
from os.path import exists


def start_mind(config=None, bus=None):
    # server
    config = config or {}
    host = config.get("host", "0.0.0.0")
    port = config.get("port", DEFAULT_PORT)
    # TODO non-ssl support
    use_ssl = config.get("ssl", USE_SSL)
    max_connections = config.get("max_connections", -1)
    address = u"wss://" + str(host) + u":" + str(port)
    cert = config.get("cert_file", DEFAULT_SSL_CRT)
    key = config.get("key_file", DEFAULT_SSL_KEY)

    factory = JarbasMind(bus=bus)
    factory.protocol = JarbasMindProtocol
    if max_connections >= 0:
        factory.setProtocolOptions(maxConnections=max_connections)

    if not exists(key) or not exists(cert):
        LOG.warning("ssl keys dont exist, creating self signed")
        name = key.split("/")[-1].replace(".key", "")
        create_self_signed_cert(CERTS_PATH, name)
        cert = CERTS_PATH + "/" + name + ".crt"
        key = CERTS_PATH + "/" + name + ".key"
        LOG.info("key created at: " + key)
        LOG.info("crt created at: " + cert)
        # update config with new keys
        config["cert_file"] = cert
        config["key_file"] = key
        # factory.config_update({"mind": config}, True)

    # SSL server context: load server key and certificate
    contextFactory = ssl.DefaultOpenSSLContextFactory(key, cert)

    reactor.listenSSL(port, factory, contextFactory)
    print("Starting mind: ", address)
    reactor.run()


if __name__ == '__main__':
    start_mind()



import os

from setuptools import setup

setup(
    name='jarbas_hive_mind',
    version='0.3.5',
    packages=['jarbas_hive_mind',
              'jarbas_hive_mind.minds',
              'jarbas_hive_mind.minds.simple',
              'jarbas_hive_mind.nodes',
              'jarbas_hive_mind.nodes.flask',
              'jarbas_hive_mind.utils',
              'jarbas_hive_mind.utils.messagebus',
              'jarbas_hive_mind.drones',
              'jarbas_hive_mind.drones.simple',
              'jarbas_hive_mind.bridges',
              'jarbas_hive_mind.bridges.hackchat',
              'jarbas_hive_mind.bridges.facebook',
              'jarbas_hive_mind.bridges.twitch',
              'jarbas_hive_mind.bridges.twitter',
              'jarbas_hive_mind.database',
              'jarbas_hive_mind.terminals',
              'jarbas_hive_mind.terminals.https',
              'jarbas_hive_mind.terminals.cli',
              'jarbas_hive_mind.terminals.basic_web',
              'jarbas_hive_mind.terminals.speech',
              'jarbas_hive_mind.terminals.speech.stt',
              'jarbas_hive_mind.terminals.speech.recognizer',
              'jarbas_hive_mind.terminals.webchat'],
    include_package_data=True,
    install_requires=["pyopenssl",
                      "service_identity",
                      "flask",
                      "flask-sslify",
                      "yagmail",
                      "autobahn",
                      "twisted",
                      "sqlalchemy",
                      "lingua_franca",
                      "requests",
                      "python-dateutil==2.6.0",
                      "tornado==4.5.3",
                      "pyee==5.0.0",
                      "websocket-client==0.54.0"],
    url='https://github.com/JarbasAl/hive_mind',
    license='MIT',
    author='jarbasAI',
    author_email='jarbasai@mailfence.com',
    description='Networking utilities for mycroft core'
)



import logging
from threading import Thread

from jarbas_hive_mind.nodes.flask.base import get_app, noindex, donation, requires_auth, request, nice_json, start
from jarbas_hive_mind.utils.messagebus.message import Message
from jarbas_hive_mind.utils.messagebus.ws import WebsocketClient

platform = "FlaskHiveNodev0.1"
LOG = logging.getLogger(platform)
LOG.setLevel("INFO")

app = get_app()
ws = None
answers = {}
users_on_hold = {}
timeout = 60


def connect():
    global ws
    ws.run_forever()


@app.route("/ask/<lang>/<utterance>", methods=['PUT', 'GET'])
@noindex
@donation
@requires_auth
def ask(utterance, lang="en-us"):
    global users_on_hold, answers
    ip = request.remote_addr
    user = request.headers["Authorization"]
    data = {"utterances": [utterance], "lang": lang}
    user_id = str(ip) + ":" + str(user)
    LOG.info("[IP] " + ip + " [UTTERANCE] " + utterance)
    context = {"source": user_id,
               "destination": "skills",
               "platform": platform}
    message = Message("recognizer_loop:utterance", data, context)
    # clean prev answer # TODO figure out how to answer both, timestamps?
    if user_id in answers:
        answers.pop(user_id)
    users_on_hold[user_id] = False
    answers[user_id] = None
    # emit message
    ws.emit(message)
    result = {"status": "processing"}
    return nice_json(result)


@app.route("/get_answer", methods=['GET', 'PUT'])
@noindex
@donation
@requires_auth
def get_answer():
    global users_on_hold, answers
    ip = request.remote_addr
    user = request.headers["Authorization"]
    user_id = str(ip) + ":" + str(user)
    if users_on_hold.get(user_id, False):
        # if answer is ready
        LOG.info("[ANSWER READY] " + ip)
        answer = Message("speak", answers[user_id]["data"], answers[
            user_id]["context"]).serialize()
        users_on_hold.pop(user_id)
        answers.pop(user_id)
        result = {"status": "done", "answer": answer}
    else:
        LOG.info("[PROCESSING ANSWER] " + ip)
        result = {"status": "processing"}
    return nice_json(result)


@app.route("/cancel", methods=['GET', 'PUT'])
@noindex
@donation
@requires_auth
def cancel_answer():
    global users_on_hold, answers
    ip = request.remote_addr
    user = request.headers["Authorization"]
    user_id = str(ip) + ":" + str(user)
    if user_id in list(users_on_hold.keys()):
        LOG.info("[CANCEL REQUEST] " + ip)
        users_on_hold.pop(user_id)
        answers.pop(user_id)
    result = {"status": "canceled"}
    return nice_json(result)


def listener(message):
    ''' listens for speak messages and checks if we are supposed to send it to some user '''
    global users_on_hold, answers
    message.context = message.context or {}
    user = message.context.get("destination", "")
    if user in list(
            users_on_hold.keys()):  # are we waiting to answer this user?

        if answers[user] is not None:
            # update data and context
            for k in list(message.context.keys()):
                answers[user]["context"][k] = message.context[k]
            for k in list(message.data.keys()):
                # update utterance
                if k == "utterance":
                    answers[user]["data"]["utterance"] = \
                        answers[user]["data"]["utterance"] + ". " + \
                        message.data["utterance"]
                else:
                    answers[user]["data"][k] = message.data[k]
        else:
            # create answer
            answers[user] = {"data": message.data,
                             "context": message.context}


def end_wait(message):
    ''' stop capturing answers for this user '''
    global users_on_hold, answers
    user = message.context.get("destination", "")
    if user in list(users_on_hold.keys()):
        LOG.info("[REQUEST COMPLETE] " + user.split(":")[0])
        # mark as answered
        users_on_hold[user] = True
        # process possible failure scenarios
        context = {}
        if message.type == "complete_intent_failure":

            answers[user] = {"data": {"utterance": "does not compute"},
                             "context": context}
        # no answer but end of handler
        elif answers[user] is None:
            answers[user] = {"data": {"utterance": "something went wrong, "
                                                   "ask me later"},
                             "context": context}


def launch(config=None):
    global app, ws, timeout
    config = config or {}
    # connect to internal mycroft
    ws = WebsocketClient()
    ws.on("mycroft.skill.handler.complete", end_wait)
    ws.on("complete_intent_failure", end_wait)
    ws.on("speak", listener)
    event_thread = Thread(target=connect)
    event_thread.setDaemon(True)
    event_thread.start()
    port = config.get("port", 6712)
    timeout = config.get("timeout", timeout)
    start(app, port)


if __name__ == "__main__":
    launch()



import json
import logging
import sys
from threading import Thread

from autobahn.twisted.websocket import WebSocketClientFactory, \
    WebSocketClientProtocol
from twisted.internet import reactor, ssl
from twisted.internet.protocol import ReconnectingClientFactory

platform = "JarbasCliTerminalv0.1"
logger = logging.getLogger(platform)
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel("INFO")


class JarbasCliTerminalProtocol(WebSocketClientProtocol):

    def onConnect(self, response):
        logger.info("[INFO] Server connected: {0}".format(response.peer))
        self.factory.client = self
        self.factory.status = "connected"

    def onOpen(self):
        logger.info("[INFO] WebSocket connection open. ")
        self.input_loop = Thread(target=self.get_cli_input)
        self.input_loop.setDaemon(True)
        self.input_loop.start()

    def onMessage(self, payload, isBinary):
        if not isBinary:
            payload = payload.decode("utf-8")
            msg = json.loads(payload)
            if msg.get("type", "") == "speak":
                utterance = msg["data"]["utterance"]
                logger.info("[OUTPUT] " + utterance)
            elif msg.get("type", "") == "hive.complete_intent_failure":
                logger.error("[ERROR] complete intent failure")
        else:
            pass

    def onClose(self, wasClean, code, reason):
        logger.info("[INFO] WebSocket connection closed: {0}".format(reason))
        self.factory.client = None
        self.factory.status = "disconnected"
        if "Internalservererror:InvalidAPIkey" in reason:
            logger.error("[ERROR] invalid user:key provided")
            raise ConnectionAbortedError("invalid user:key provided")

    # cli input thread
    def get_cli_input(self):
        logger.info("[INFO] waiting for input")
        while True:
            line = input("")
            msg = {"data": {"utterances": [line], "lang": "en-us"},
                   "type": "recognizer_loop:utterance",
                   "context": {"source": self.peer,
                               "destination": "hive_mind",
                               "platform": platform}}
            msg = json.dumps(msg)
            msg = bytes(msg, encoding="utf-8")
            self.sendMessage(msg, False)


class JarbasCliTerminal(WebSocketClientFactory, ReconnectingClientFactory):
    protocol = JarbasCliTerminalProtocol

    def __init__(self, *args, **kwargs):
        super(JarbasCliTerminal, self).__init__(*args, **kwargs)
        self.status = "disconnected"
        self.client = None

    # websocket handlers
    def clientConnectionFailed(self, connector, reason):
        logger.info("[INFO] Client connection failed: " + str(
            reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)

    def clientConnectionLost(self, connector, reason):
        logger.info("[INFO] Client connection lost: " + str(
            reason) + " .. retrying ..")
        self.status = "disconnected"
        self.retry(connector)

